import collections
import itertools
import torch
import warnings

__all__ = ['Lookahead']

#NOTE: modified from https://github.com/alphadl/lookahead.pytorch

#TODO: create factory optimizer
class Lookahead(torch.optim.Optimizer):
    """Implements the Lookahead algorithm.

    - **Paper**: [Lookahead Optimizer: k steps forward, 1 step back](https://arxiv.org/pdf/1907.08610.pdf)
    - **Implementation**: [GitHub @ alphadl](https://github.com/alphadl/lookahead.pytorch)

    """
    def __init__(self,
        optimizer: torch.optim.Optimizer,
        k: int=5,
        alpha: float=0.5
    ):
        self.optimizer = optimizer
        self.k = k
        self.alpha = alpha
        self.param_groups = self.optimizer.param_groups
        self.state = collections.defaultdict(dict)
        self.fast_state = self.optimizer.state
        for group in self.param_groups:
            group["counter"] = 0
    
    def update(self, group):
        for fast in group["params"]:
            param_state = self.state[fast]
            if "slow_param" not in param_state:
                param_state["slow_param"] = torch.zeros_like(fast.data)
                param_state["slow_param"].copy_(fast.data)
            slow = param_state["slow_param"]
            slow += (fast.data - slow) * self.alpha
            fast.data.copy_(slow)
    
    def update_lookahead(self):
        for group in self.param_groups:
            self.update(group)

    def step(self, closure=None):
        loss = self.optimizer.step(closure)
        for group in self.param_groups:
            if group["counter"] == 0:
                self.update(group)
            group["counter"] += 1
            if group["counter"] >= self.k:
                group["counter"] = 0
        return loss

    def state_dict(self):
        fast_state_dict = self.optimizer.state_dict()
        slow_state = {
            (id(k) if isinstance(k, torch.Tensor) else k): v
            for k, v in self.state.items()
        }
        fast_state = fast_state_dict["state"]
        param_groups = fast_state_dict["param_groups"]
        return {
            "fast_state": fast_state,
            "slow_state": slow_state,
            "param_groups": param_groups,
        }

    def load_state_dict(self, state_dict):
        slow_state_dict = {
            "state": state_dict["slow_state"],
            "param_groups": state_dict["param_groups"],
        }
        fast_state_dict = {
            "state": state_dict["fast_state"],
            "param_groups": state_dict["param_groups"],
        }
        super(Lookahead, self).load_state_dict(slow_state_dict)
        self.optimizer.load_state_dict(fast_state_dict)
        self.fast_state = self.optimizer.state

    def add_param_group(self, param_group):
        param_group["counter"] = 0
        self.optimizer.add_param_group(param_group)



import moai.checkpoint.lightning as mickpt
import moai.log.lightning as milog

import pytorch_lightning
import hydra.utils as hyu
import omegaconf.omegaconf
import typing

__all__ = ["LightningTester"]

class LightningTester(pytorch_lightning.Trainer):
    def __init__(self,
        logging:                    omegaconf.DictConfig=None,
        callbacks:                  omegaconf.DictConfig=None,
        default_root_dir:           typing.Optional[str]=None,
        process_position:           int=0,
        num_nodes:                  int=1,
        num_processes:              int=1,
        gpus:                       typing.Optional[typing.Union[typing.List[int], str, int]]=None,
        tpu_cores:                  typing.Optional[typing.Union[typing.List[int], str, int]]=None,
        log_gpu_memory:             typing.Optional[str]=None,
        max_steps:                  typing.Optional[int]=None,
        min_steps:                  typing.Optional[int]=None,
        limit_test_batches:         typing.Union[int, float]=1.0,
        limit_predict_batches:      typing.Union[int, float]=1.0,
        accelerator:                typing.Optional[typing.Union[str, pytorch_lightning.accelerators.Accelerator]]=None,
        strategy:                   typing.Optional[str]=None,
        sync_batchnorm:             bool=False,
        precision:                  int=32,
        enable_model_summary:       bool=True,
        weights_summary:            typing.Optional[str]='full',
        weights_save_path:          typing.Optional[str]=None,        
        resume_from_checkpoint:     typing.Optional[str]=None,
        profiler:                   typing.Optional[typing.Union[pytorch_lightning.profiler.BaseProfiler, bool, str]]=None,
        benchmark:                  bool=False,
        deterministic:              bool=True,
        replace_sampler_ddp:        bool=True,
        prepare_data_per_node:      bool=True,
        plugins:                    typing.Optional[list]=None,
        amp_backend:                str='native',
        amp_level:                  str='O2',
        **kwargs
    ):
        logger = hyu.instantiate(logging)\
            if logging is not None else milog.NoOp()
        pytl_callbacks = [hyu.instantiate(c) for c in callbacks.values()]\
            if callbacks is not None else []                
        super(LightningTester, self).__init__(
            logger=logger,
            checkpoint_callback=False,
            callbacks=pytl_callbacks,
            default_root_dir=None if not default_root_dir else default_root_dir,
            gradient_clip_val=0.0,
            process_position=process_position,
            num_nodes=num_nodes,
            gpus=gpus,
            auto_select_gpus=False,
            tpu_cores=tpu_cores,
            log_gpu_memory=log_gpu_memory,
            progress_bar_refresh_rate=1,
            overfit_batches=0.0,
            track_grad_norm=-1,
            check_val_every_n_epoch=1,
            fast_dev_run=False,
            accumulate_grad_batches=1,
            max_epochs=1,
            min_epochs=1,
            max_steps=max_steps,
            min_steps=min_steps,
            max_time=None, #NOTE @PTL1.5 #TODO: check if needed
            limit_train_batches=1.0,
            limit_val_batches=1.0,
            limit_test_batches=limit_test_batches,
            limit_predict_batches=limit_predict_batches, #NOTE: @PTL1.5
            val_check_interval=1,
            flush_logs_every_n_steps=1,
            log_every_n_steps=1,
            accelerator=accelerator,
            strategy=strategy, #NOTE: @PTL1.5
            sync_batchnorm=sync_batchnorm,
            precision=precision,
            enable_model_summary=enable_model_summary,
            weights_summary=weights_summary,
            weights_save_path=weights_save_path,
            num_sanity_val_steps=0,
            resume_from_checkpoint=resume_from_checkpoint,
            profiler=profiler,
            benchmark=benchmark,
            deterministic=deterministic,
            reload_dataloaders_every_epoch=False,
            reload_dataloaders_every_n_epochs=False,
            auto_lr_find=False,
            detect_anomaly=False, #NOTE: @PTL1.5
            replace_sampler_ddp=replace_sampler_ddp,
            terminate_on_nan=False,
            auto_scale_batch_size=False,
            prepare_data_per_node=prepare_data_per_node,
            plugins=plugins,
            amp_backend=amp_backend,            
            amp_level=amp_level,
            #NOTE: @PTL1.5 automatic_optimization=False,
            move_metrics_to_cpu=False,
            multiple_trainloader_mode="max_size_cycle",
            stochastic_weight_avg=False,
            num_processes=num_processes, #NOTE: @PTL1.5 fix
            **kwargs
        )

    def run(self, model):
        return self.test(model, verbose=False)


import logging
import typing

def assert_choices(
    logger:         logging.Logger,
    name:    str,
    choice:         typing.Any,
    choices:        typing.Sequence[typing.Any],
) -> None:
    if choice not in choices:
        logger.error(f"The value given for {name} ({choice}) is invalid,"
            f" acceptable values are: {choices}"
        )

def ensure_choices(
    logger:         logging.Logger,
    name:    str,
    choice:         typing.Any,
    choices:        typing.Sequence[typing.Any],
) -> None:
    if choice not in choices:
        logger.error(f"The value given for {name} ({choice}) is invalid,"
            f" acceptable values are: {choices}"
        )
    return choice



import torch
import functools
import typing

__all__ = [
    "Upsample2d",
]

class Upsample2d(torch.nn.Module):
    def __init__(self,
        resolution: typing.Sequence[int]=None,
        scale: float=2.0,        
        mode: str="bilinear",
    ):
        super(Upsample2d, self).__init__()
        if resolution:
            self.upsample = functools.partial(torch.nn.functional.interpolate,
                size=tuple(resolution),
                mode=mode
            )
        else:
            self.upsample = functools.partial(torch.nn.functional.interpolate,
                scale_factor=scale,
                mode=mode
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.upsample(x)


import torch
import logging

log = logging.getLogger(__name__)

__all__ = ["NoOp"]

class NoOp(torch.utils.data.Dataset):
    def __init__(self,        
        dataset: torch.utils.data.Dataset
    ):
        super(NoOp, self).__init__()
        self.inner = dataset
        log.info(f"No data augmentation being used.")

    def __len__(self):
        return len(self.inner)

    def __getitem__(self, index: int):
        return self.inner[index] # inner dataset item


from moai.networks.lightning.factory.autoencoder import Autoencoder
from moai.networks.lightning.factory.stacked_hourglass import StackedHourglass
from moai.networks.lightning.factory.vae import VariationalAutoencoder
from moai.networks.lightning.factory.multibranch import MultiBranch
from moai.networks.lightning.factory.hrnet import HRNet
from moai.networks.lightning.factory.cascade import Cascade

__all__ = [
    "Autoencoder",
    "HRNet",
    "MultiBranch",
    "StackedHourglass",
    "VariationalAutoencoder",
    "Cascade"
]


from moai.utils.arguments import assert_numeric
from collections.abc import Callable
from pyviz3d.visualizer import Visualizer

import pyviz3d
import typing
import torch
import logging
import colour
import shutil
import numpy as np
import math
import os
import json

log = logging.getLogger(__name__)

__all__ = ['PointCloud']

class PointCloud(Callable):    
    def __init__(self,
        points:            typing.Union[str, typing.Sequence[str]],
        color:             typing.Union[str, typing.Sequence[str]],
        batch_percentage:   float=1.0,
        name:               str="default",
        port:               int=8098,
    ):
        self.name, self.port = name, port
        self.vertices = [points] if isinstance(points, str) else list(points)
        self.colors = [colour.Color(color)] if isinstance(color, str) else list(colour.Color(c) for c in color)
        self.batch_percentage = batch_percentage
        assert_numeric(log, 'batch percentage', batch_percentage, 0.0, 1.0)
        self.visualizer = Visualizer()
        # os.makedirs('pyviz3d', exist_ok=True)
        directory_source = os.path.realpath(
            os.path.join(os.path.dirname(pyviz3d.__file__), "src")
        )
        self.directory_destination = os.path.join(os.getcwd(), 'pyviz3d')
        shutil.copytree(directory_source, self.directory_destination)
        log.warning(f"PyViz3D visualization enabled. Run `python -m http.server {port}` @ {os.path.join(os.getcwd(), 'pyviz3d')} and then navigate to localhost:{port}.")

    def _save(self):
        # Assemble binary data files
        nodes_dict = {}
        for name, e in self.visualizer.elements.items():
            binary_file_path = os.path.join(self.directory_destination, name + ".bin")
            nodes_dict[name] = e.get_properties(name + ".bin")
            e.write_binary(binary_file_path)
        # Write json file containing all scene elements
        json_file = os.path.join(self.directory_destination, "nodes.json")
        with open(json_file, "w") as outfile:
            json.dump(nodes_dict, outfile)

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for v, c in zip(self.vertices, self.colors):
            vertices = tensors[v].detach().cpu().numpy()
            b = math.ceil(self.batch_percentage * vertices.shape[0])
            for i in range(b):                
                colors = (np.array(c.rgb) * 255.0).astype(np.uint8)
                colors = colors[np.newaxis, ...].repeat(vertices.shape[1], 0)
                self.visualizer.add_points(f"{v}_{i}", vertices[i], colors, point_size=1.0)
        self._save()

'''
import http.server
from http.server import HTTPServer, BaseHTTPRequestHandler
import socketserver

PORT = 8080

Handler = http.server.SimpleHTTPRequestHandler

Handler.extensions_map={
        '.manifest': 'text/cache-manifest',
    '.html': 'text/html',
        '.png': 'image/png',
    '.jpg': 'image/jpg',
    '.svg':	'image/svg+xml',
    '.css':	'text/css',
    '.js':	'text/javascript',
    '': 'application/octet-stream', # Default
    }

httpd = socketserver.TCPServer(("", PORT), Handler)

print("serving at port", PORT)
httpd.serve_forever()
'''


import torch

#NOTE: code from: https://github.com/facebookresearch/pytorch3d/blob/master/pytorch3d/transforms/so3.py  

__all__ = ["AngleError"]

def _angular_error(
    gt:             torch.Tensor,
    pred:           torch.Tensor,
    radians:        bool
): 
    relative = gt @ torch.transpose(pred, -2, -1)
    trace = relative[:, 0, 0] + relative[:, 1, 1] + relative[:, 2, 2] 
    trace = torch.clamp(trace, -1.0, 3.0)
    phi = 0.5 * (trace - 1.0)
    return phi.acos() if radians else torch.rad2deg(phi.acos())

class AngleError(torch.nn.Module):
    def __init__(self,
        radians: bool = True,
    ):
        super(AngleError, self).__init__()
        self.radians = radians

    def forward(self,
        gt:             torch.Tensor,
        pred:           torch.Tensor,        
    ) -> torch.Tensor:
        return _angular_error(gt, pred, self.radians).mean()


from moai.monads.geometry.project import Projection
from moai.monads.geometry.transform_points import Transformation
from moai.monads.geometry.transform_ops import (
    Relative,
    Inverse,
    Transpose
)
from moai.monads.geometry.camera import WeakPerspective as WeakPerspectiveCamera
from moai.monads.geometry.rotate import Rotate

__all__ = [
    "Projection",
    "Transformation",
    "Relative",
    "Inverse",
    "Transpose",
    "WeakPerspectiveCamera",
    "Rotate",
]


from moai.monads.utils.common import dim_list

import torch

class RMSE(torch.nn.Module):
    def __init__(self):
        super(RMSE, self).__init__()

    def forward(self,
        gt:         torch.Tensor,
        pred:       torch.Tensor,
        weights:    torch.Tensor=None,
        mask:       torch.Tensor=None,
    ) -> torch.Tensor:
        diff_sq = (gt - pred) ** 2
        if weights is not None:
            diff_sq = diff_sq * weights
        if mask is not None:
            diff_sq = diff_sq[mask]
        if weights is None:
            return torch.mean(torch.sqrt(torch.mean(diff_sq, dim=dim_list(gt))))
        else:
            diff_sq_sum = torch.sum(diff_sq, dim=dim_list(gt))
            diff_w_sum = torch.sum(weights, dim=dim_list(gt)) # + 1e-18
            return torch.mean(torch.sqrt(diff_sq_sum / diff_w_sum))


import logging
import typing

def ensure_numeric_list(
    arg: typing.Union[int, float, typing.Sequence[int], typing.Sequence[float]],
) -> typing.List[typing.Union[int, float]]:
    return [arg] if type(arg) is int or type(arg) is float else arg

def ensure_string_list(
    arg: typing.Union[str, typing.Sequence[str]],
) -> typing.List[str]:
    return [arg] if type(arg) is str else arg

def assert_sequence_size(
    logger:         logging.Logger,
    name:           str,
    sequence:       typing.Sequence[typing.Any],
    max_size:       int,
    min_size:       int=1,    
) -> None:
    length = len(sequence)
    if length < min_size:
        logger.error(f"List {name} (length: {length}) should have at least {min_size} elements.")
    if length > max_size:
        logger.error(f"List {name} (length: {length}) should have at most {max_size} elements.")


import moai.utils.engine as mieng

import omegaconf.omegaconf

__all__ = ['Statistics']

class Statistics(mieng.Collection):
    def __init__(self, 
        calculators: omegaconf.DictConfig,        
    ):
        super().__init__(items=calculators, name="calculators")


from moai.monads.distribution.center_of_mass import CenterOfMass
from moai.monads.distribution.coordinate_decoding import CoordinateDecoding
from moai.monads.distribution.taylor import Taylor

__all__ = [
    "CenterOfMass",
    "CoordinateDecoding",
    "Taylor"
]


import torch
import logging
import kornia as kn

log = logging.getLogger(__name__)

__all__ = [
    "Convert",
    "AxisAngle",
]

#TODO: make_from, convert_to
class AxisAngle(torch.nn.Module):
    def __init__(self,
        to:    str = "mat", #"quat" , "ortho6d", "euler_angles" , etc.        
    ):
        super().__init__()
        if to == "mat":
            self.convert_func = kn.geometry.conversions.angle_axis_to_rotation_matrix
        elif to == "quat":
            #(x, y, z, w) format of quaternion
            self.convert_func = kn.geometry.conversions.angle_axis_to_quaternion
        else:
            log.error(f"The selected rotational convertion {to} is not supported. Please enter a valid one to continue.")

    def forward(self,angle_axis: torch.Tensor) -> torch.Tensor:
        return self.convert_func(angle_axis)

class Convert(torch.nn.Module):
    __MAPPING__ = {
        'aa': AxisAngle,        
    }

    def __init__(self,
        from_to:       str,
    ):
        super().__init__()
        f, t = from_to.split('_')
        self.converter = Convert.__MAPPING__[f](to=t)

    def forward(self, rotation: torch.Tensor) -> torch.Tensor:
        return self.converter(rotation)


from moai.serve.handlers.image_raw import ImageRaw
from moai.serve.handlers.image_file import (
    ImageFileInput,
    ImageFileOutput, 
)

__all__ = [
    'ImageRaw',
    'ImageFileInput',
    'ImageFileOutput',
]


import tablib
import toolz
import pytorch_lightning
import typing
import logging 

log = logging.getLogger(__name__)

__all__ = ["Tabular"]

class Tabular(pytorch_lightning.loggers.base.LightningLoggerBase):
    def __init__(self,
        name:           str="default",
        version:        int=0,
    ):
        super(Tabular, self).__init__()   
        self._version = version
        self.file_name = name
        self.train_logs = tablib.Dataset()
        self.val_logs = tablib.Dataset()
        self.test_logs = tablib.Dataset()
        self.train_headers_written = False
        self.val_headers_written = False
        self.test_headers_written = False

    @property
    def name(self) -> str:
        return self.file_name

    def _append_train_losses(self,
        metrics:            typing.Dict[str, typing.Any],
        epoch:              int,
        step:               int,
    ) -> None:
        loss = metrics['total_loss']
        train_metrics = toolz.dissoc(metrics, 'train', 'epoch', 'total_loss')
        if self.train_logs.headers is None and not self.train_headers_written:
            self.train_logs.headers = list(toolz.concat([
                [str('epoch'), str('iteration'), str('total_loss')],
                [k for k in train_metrics.keys()]
            ]))
        self.train_logs.append(list(
            toolz.concat([
                [epoch, step, loss],
                train_metrics.values()
            ])
        ))
        
    def _append_val_loss(self, 
        metrics:        typing.Dict[str, typing.Any],
        epoch:          int,
        step:           int,
    ) -> None:
        if self.val_logs.headers is None and not self.val_headers_written:
            self.val_logs.headers = list(toolz.concat([
                [str('epoch'), str('iteration')],
                [k for k in metrics.keys()]
            ]))
            self.val_headers = self.val_logs.headers
        self.val_logs.append(list(
            toolz.concat([
                [epoch, step],
                [metrics[k] for k in self.val_headers if k in metrics]
            ])
        ))

    def _append_test_metrics(self, 
        metrics:        typing.Dict[str, typing.Any],
        step:           int,
    ) -> None:
        if self.test_logs.headers is None and not self.test_headers_written:
            self.test_logs.headers = list(toolz.concat([
                [str('iteration')],
                [k for k in metrics.keys()]
            ]))
            self.test_headers = self.test_logs.headers
        self.test_logs.append(list(
            toolz.concat([
                [step],
                [metrics[k] for k in self.test_headers if k in metrics]
            ])
        ))

    @pytorch_lightning.loggers.base.rank_zero_only
    def log_metrics(self, 
        metrics:        typing.Dict[str, typing.Any],
        step:           int
    ) -> None:
        train_metrics = toolz.keymap(lambda k: k.replace('train_', ''), 
            toolz.keyfilter(lambda k: k.startswith('train_'), metrics)
        )
        val_metrics = toolz.keymap(lambda k: k.replace('val_', ''), 
            toolz.keyfilter(lambda k: k.startswith('val_'), metrics)
        )
        test_metrics = toolz.keymap(lambda k: k.replace('test_', '').replace('/epoch_0', ''), 
            toolz.keyfilter(lambda k: k.startswith('test_'), metrics)
        )
        if train_metrics:
            self._append_train_losses(
                toolz.assoc(train_metrics, 'total_loss', metrics['total_loss']), 
                metrics['epoch'],
                step
            )
        elif test_metrics:            
            self._append_test_metrics(test_metrics, step)
            return
        if val_metrics:
            self._append_val_loss(val_metrics, metrics['epoch'], step)

    @pytorch_lightning.loggers.base.rank_zero_only
    def log_hyperparams(self,
        params: typing.Dict[str, typing.Any] #TODO or namespace object ?
    ) -> None:
        """Record hyperparameters
            :param params: argparse.Namespace containing the hyperparameters
        """
        data = tablib.Dataset()
        data.headers = [k for k in dict(params).keys()]
        data.append([v for v in dict(params).values()])
        with open(self.name + "_hparams.yaml", 'w') as f:
            f.write(data.export('yaml'))

    @pytorch_lightning.loggers.base.rank_zero_only
    def save(self) -> None:
        if self.train_logs.height:
            """Save train log data"""
            with open(self.name + "_train.csv", 'a', newline='') as f:
                f.write(self.train_logs.export('csv'))
            if not self.train_headers_written and self.train_logs.headers is not None:
                self.train_headers_written = True
            self.train_logs.wipe()        
        if self.val_logs.height:
            """Save val log data"""
            with open(self.name + "_val.csv", 'a', newline='') as f:
                f.write(self.val_logs.export('csv'))
            if not self.val_headers_written and self.val_logs.headers is not None:
                self.val_headers_written = True
            self.val_logs.wipe()
        if self.test_logs.height:
            """Save val log data"""
            with open(self.name + "_test.csv", 'a', newline='') as f:
                f.write(self.test_logs.export('csv'))
            if not self.test_headers_written and self.test_logs.headers is not None:
                self.test_headers_written = True
            self.test_logs.wipe()

    @pytorch_lightning.loggers.base.rank_zero_only
    def finalize(self, 
        status: str
    ) -> None:
        """Do any processing that is necessary to finalize an experiment
            :param status: Status that the experiment finished with (e.g. success, failed, aborted)
        """
        self.save()
        self.close()

    @property
    def rank(self) -> int:
        """
            Process rank. In general, metrics should only be logged by the process
                with rank 0
        """
        return self._rank

    @rank.setter
    def rank(self, value: int) -> None:
        """Set the process rank"""
        self._rank = value

    @property
    def version(self) -> int:
        """Return the experiment version"""
        return self._version

    @property
    def experiment(self) -> typing.Any:
        return self.name


import os
import omegaconf as oconf
import deepdiff as dd
import hydra
import logging
import itertools
import rich
from rich.console import Console
from rich import box

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

def get_conf(path: str) -> oconf.OmegaConf:
    if os.path.splitext(path)[1] != '.yaml':
        nominal_config_path = os.path.join(path, 'config_resolved.yaml')
        if os.path.exists(nominal_config_path):
            path = nominal_config_path
        elif os.path.basename(path) != '.hydra':
            path = os.path.join(path, '.hydra') #TODO: merge with overrides
            path = os.path.join(path, 'config.yaml')
    return oconf.OmegaConf.load(path)

def report_hydra(ddiff: dd.DeepDiff) -> None:
    def add_item(txt: str) -> None:
        log.debug(f"\t{txt}")
    def remove_item(txt: str) -> None:
        log.error(f"\t{txt}")
    def value_changed(txt: str) -> None:
        log.warning(f"\t{txt}")
    def key_changed(txt: str) -> None:
        log.critical(f"\t{txt}")

    def format_added_item(txt: str) -> str:
        ret = txt.replace('\'', '').\
            replace('root', '').\
            replace('[', '').\
            replace(']', '.')
        return ret[:-1] if ret.endswith('.') else ret

    def format_value_changed(txt: str) -> str:
        old = txt[1]['old_value']
        new = txt[1]['new_value']
        return txt[0].replace('\'','').\
            replace('][', '.').\
            replace('root[', '').\
            replace(']','')\
            + f': {old} -> {new}'
 
    def format_iterable_item(txt: str) -> str:
        ret = txt[:-3].replace('\'', '').\
            replace('root', '').\
            replace('[', '').\
            replace(']', '.') 
        return (ret[:-1] if ret.endswith('.') else ret) + txt[-3:]

    if 'dictionary_item_removed' in ddiff.keys():
        log.info("- Dictionary Items Removed:")
        for i in ddiff['dictionary_item_removed'].items:
            formatted = format_added_item(i)
            remove_item(formatted)

    if 'dictionary_item_added' in ddiff.keys():
        log.info("+ Dictionary Items Added:")
        for i in ddiff['dictionary_item_added'].items:
          formatted = format_added_item(i)
          add_item(formatted)

    if 'iterable_item_added' in ddiff.keys():
        log.info("+ Iterable Items Added:")
        for i in ddiff['iterable_item_added'].items():
          add_item(f'{format_iterable_item(i[0])} -> {format_iterable_item(i[1])}')
    
    if 'iterable_item_removed' in ddiff.keys():
        log.info("- Iterable Items Removed:")
        for i in ddiff['iterable_item_removed'].items():
          remove_item(f'{format_iterable_item(i[0])} -> {format_iterable_item(i[1])}')
    
    if 'type_changes' in ddiff.keys():
        log.info("~ Types Changed:")
        for i in ddiff['type_changes'].items():
          key_changed(i)

    if 'values_changed' in ddiff.keys():
        log.info("~ Values Changed:")
        for i in ddiff['values_changed'].items():
          value_changed(format_value_changed(i))

    if 'repetition_change' in ddiff.keys():
        log.info("~ Repetition Changed:")
        for i in ddiff['repetition_change'].items():
          value_changed(i)

    if 'set_item_added' in ddiff.keys():    
        log.info("+ Set Item Added:")
        for i in ddiff['set_item_added'].items():
          add_item(i)

    if 'set_item_removed' in ddiff.keys():    
        log.info("- Set Item Removed:")
        for i in ddiff['set_item_removed'].items():
          remove_item(i)

def report_rich(ddiff: dd.DeepDiff, console: Console) -> None:
    key_table = rich.table.Table(box=box.SIMPLE_HEAVY, border_style="bold yellow")
    key_table.add_column(":heavy_check_mark: [bold green] Added", justify='center')
    key_table.add_column(":x: [bold red] Removed", justify='center')

    collection_table = rich.table.Table(box=box.SIMPLE_HEAVY, border_style="bold blue")
    collection_table.add_column(":key: [bold blue] Key", justify='center')
    collection_table.add_column(":x: [bold red] From", justify='center')
    collection_table.add_column(":heavy_check_mark: [bold green] To", justify='center')

    value_table = rich.table.Table(box=box.SIMPLE_HEAVY, border_style='bold cyan')
    value_table.add_column(":key: [bold cyan] Key", justify='center')
    value_table.add_column(":x: [bold red] From", justify='center')
    value_table.add_column(":heavy_check_mark: [bold green] To", justify='center')

    results = {
        'item': {
            'added': [], 'removed': [],
        },
        'collection': {
            'added': {}, 'removed': {},
        },
        'changed': {
            'new': [], 'key': [], 'old': [], 
        },
        'repetition': {
            'added': [], 'removed': [],
        }
    }

    def format_added_item(txt: str) -> str:
        ret = txt.replace('\'', '').\
            replace('root', '').\
            replace('[', '').\
            replace(']', '.')
        return ret[:-1] if ret.endswith('.') else ret

    def format_value_changed(txt: str) -> str:
        old = txt[1]['old_value']
        new = txt[1]['new_value']
        return txt[0].replace('\'','').\
            replace('][', '.').\
            replace('root[', '').\
            replace(']','')\
            + f': {old} -> {new}'
 
    def format_iterable_item(txt: str) -> str:
        ret = txt[:-3].replace('\'', '').\
            replace('root', '').\
            replace('[', '').\
            replace(']', '.') 
        return (ret[:-1] if ret.endswith('.') else ret) + txt[-3:]

    if 'dictionary_item_removed' in ddiff.keys():
        for i in ddiff['dictionary_item_removed'].items:
            formatted = format_added_item(i)
            results['item']['removed'].append(formatted)

    if 'dictionary_item_added' in ddiff.keys():
        for i in ddiff['dictionary_item_added'].items:
          formatted = format_added_item(i)
          results['item']['added'].append(formatted)

    if 'iterable_item_added' in ddiff.keys():
        for i in ddiff['iterable_item_added'].items():
          key = format_iterable_item(i[0])
          value = format_iterable_item(i[1] if isinstance(i[1], str) else str(i[1]))
          results['collection']['added'][key] = value
    
    if 'iterable_item_removed' in ddiff.keys():
        for i in ddiff['iterable_item_removed'].items():
          key = format_iterable_item(i[0])
          value = format_iterable_item(i[1] if isinstance(i[1], str) else str(i[1]))
          results['collection']['removed'][key] = value
    
    if 'type_changes' in ddiff.keys():
        for i in ddiff['type_changes'].items():
            if isinstance(i, tuple):                
                results['changed']['key'].append(format_added_item(i[0]))
                results['changed']['new'].append(i[1]['new_value'])
                results['changed']['old'].append(i[1]['old_value'])
            else:
                results['changed']['key'].append(i)

    if 'values_changed' in ddiff.keys():
        for i in ddiff['values_changed'].items():            
            results['changed']['key'].append(format_added_item(i[0]))
            results['changed']['new'].append(i[1]['new_value'])
            results['changed']['old'].append(i[1]['old_value'])

    if 'repetition_change' in ddiff.keys():
        for i in ddiff['repetition_change'].items():
            pass

    if 'set_item_added' in ddiff.keys():    
        for i in ddiff['set_item_added'].items():
            pass

    if 'set_item_removed' in ddiff.keys():    
        for i in ddiff['set_item_removed'].items():
            pass

    if results['item']['added']:
        console.rule("Keys Diffs", style="bold yellow")
        for i, (a, r) in enumerate(itertools.zip_longest(
            results['item']['added'], 
            results['item']['removed'])
        ):
            key_table.add_row(
                f'[bold green] {a}' if a is not None else '',
                f'[bold red] {r}' if r is not None else '',
                end_section=(i == len(results['item']['added']) - 1)
            )
        console.print(key_table, justify='center')

    collection_added = set(results['collection']['added'].keys())
    collection_removed = set(results['collection']['removed'].keys())
    collection_diffs = collection_added.intersection(collection_removed)
    collection_removed = collection_removed - collection_diffs
    collection_added = collection_added - collection_diffs
    if collection_diffs:
        console.rule("Collections Diffs", style="bold blue")
        for k in collection_diffs:
            a = results['collection']['added'][k]
            r = results['collection']['removed'][k]
            collection_table.add_row(
                f'[bold blue] {k}',
                f'[bold red] {r}',
                f'[bold green] {a}'
            )
    if collection_added:
        for k in collection_added:
            a = results['collection']['added'][k]            
            collection_table.add_row(
                f'[bold blue] {k}',
                '',
                f'[bold green] {a}'
            )
    if collection_removed:
        for k in collection_removed:
            r = results['collection']['removed'][k]            
            collection_table.add_row(
                f'[bold blue] {k}',
                f'[bold red] {r}'
                '',                
            )
    if collection_added or collection_removed or collection_diffs:
        console.print(collection_table, justify='center')
    
    if results['changed']['key']:
        console.rule("[bold syan] Value Diffs", style="bold cyan")
        for k, n, o in zip(        
            results['changed']['key'],
            results['changed']['new'], 
            results['changed']['old'], 
        ):
            value_table.add_row(
                f'[bold cyan] {k}',
                f'[bold red] {o}',
                f'[bold green] {n}',                
            )
        console.print(value_table, justify='center')
        
def replace_bool(conf: dict) -> None:
    for k, v in conf.items():
        if isinstance(v, bool):
            conf[k] = str(v)
        if isinstance(v, oconf.ListConfig):
            for i, b in enumerate(v):
                if isinstance(b, bool):
                    conf[k][i] = str(b)
        if isinstance(v, oconf.DictConfig):
            replace_bool(v)

#NOTE: https://github.com/ikatyang/emoji-cheat-sheet 

@hydra.main(config_name="tools/diff.yaml", config_path="conf")
def diff(cfg):    
    console = Console()
    console.rule("Config Diff", style="magenta")
    from_style = rich.style.Style(frame=True)
    console.line()
    console.print("[bold red]From:[/bold red] :arrow_right_hook:", cfg.lhs, style=from_style)    
    # console.rule()
    to_style = rich.style.Style(frame=True)
    console.print("[bold green]To:[/bold green] ", cfg.rhs, justify='right', style=to_style)
    console.line()
    # console.rule()
    with console.status("Loading configs ...", spinner='aesthetic') as status:
        lhs = get_conf(cfg.lhs)
        rhs = get_conf(cfg.rhs)
        replace_bool(lhs) # bool is unhashable
        replace_bool(rhs) # bool is unhashable
    with console.status("Calculating diffs ...", spinner='aesthetic') as status:
        ddiff = dd.DeepDiff(lhs, rhs, ignore_order=True)
    # report_hydra(ddiff)
    report_rich(ddiff, console)


if __name__ == "__main__":
    # os.environ['HYDRA_FULL_ERROR'] = str(1)
    diff()



from moai.utils.torch import get_submodule

import torch
import typing
import logging

__all__ = ['NamedParameterSelector']

log = logging.getLogger(__name__)

class NamedParameterSelector(typing.Callable[[torch.nn.Module], typing.List[torch.Tensor]]):
    def __init__(self,
        keys:       typing.Sequence[str],
        force_grad: bool=True,
    ):
        self.keys = keys
        self.force_grad = force_grad

    def __call__(self, module: torch.nn.Module) -> typing.List[torch.Tensor]:
        params = []
        for k in self.keys:
            m = module
            found = True
            for s in k.split('.'): #TODO: check get_submodule in pytorch 1.10
                found = found and hasattr(m, s)
                if found:
                    m = getattr(m, s)
                else:
                    break
            if found:
                if self.force_grad:
                    m.requires_grad_(True)
                params.append(m)
            else:
                log.warning(f"Parameter {k} not found!")
        return params


from moai.parameters.optimization.optimizers.larc import LARC as Outer

import torch
import omegaconf.omegaconf
import hydra.utils as hyu
import typing

__all__ = ['LARC']

#NOTE: modified from https://github.com/alphadl/lookahead.pytorch

class LARC(object):
    """ Implements the LARC algorithm.

    - **Paper**: [Large Batch Training of Convolutional Networks](https://arxiv.org/pdf/1708.03888.pdf)
    - **Implementation**: [GitHub @ NVIDIA](https://github.com/NVIDIA/apex/blob/master/apex/parallel/LARC.py)

    """
    def __init__(self,
        parameters: typing.Iterator[torch.nn.Parameter],
        optimizers: omegaconf.DictConfig,
        trust_coefficient: float=0.02,
        clip: bool=True,
        eps: float=1e-8,
    ):
        self.optimizers = [
            Outer(
                optimizer=hyu.instantiate(optimizers[0], parameters), 
                trust_coefficient=trust_coefficient, clip=clip, eps=eps,
            )
        ]



from moai.utils.color.colorize import get_colormap, COLORMAPS
from moai.visualization.visdom.base import Base

import torch
import visdom
import numpy
import functools
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["Blend2d"]

class Blend2d(Base):
    def __init__(self,
        left:           typing.Union[str, typing.Sequence[str]],
        right:          typing.Union[str, typing.Sequence[str]],
        blending:       typing.Union[float, typing.Sequence[float]],
        colormap:       typing.Union[str, typing.Sequence[str]],
        transform:      typing.Union[str, typing.Sequence[str]],
        scale:          float=1.0,
        name:           str="default",
        ip:             str="http://localhost",
        port:           int=8097,   
    ):
        super(Blend2d, self).__init__(name, ip, port)                
        self.left = [left] if type(left) is str else list(left)
        self.right = [right] if type(right) is str else list(right)
        self.blending = [blending] if type(blending) is float else list(blending)
        self.transforms = [transform] if type(transform) is str else list(transform)
        self.colormaps = [colormap] if type(colormap) is str else list(colormap)
        self.scale = scale
        self.transform_map = {
            'none': functools.partial(self.__no_transform),
            'minmax': functools.partial(self.__minmax_normalization),
            'sum_minmax': functools.partial(self.__sum_minmax_norm)
        }
        self.colorize_map = { "none": lambda x: x }
        self.colorize_map.update(COLORMAPS)

    @property
    def name(self) -> str:
        return self.env_name
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for l, r, b, t, c in zip(self.left, self.right, self.blending, 
            self.transforms, self.colormaps):
                n = l + "_" + r
                left = tensors[l]
                left = left.cpu().detach().numpy() if left.is_cuda else left.detach().numpy()                
                self.__viz_color(self.visualizer,
                    left * b + (1.0 - b) * self.colorize_map[c](
                        self.transform_map[t](tensors[r])), n, n, self.name,
                    self.scale
                )

    @staticmethod
    def __viz_color(
        visdom: visdom.Visdom,
        array:  numpy.array,
        key:    str,
        win:    str,
        env:    str,
        scale:  float,
    ) -> None:
        if scale != 1.0:
            array = torch.nn.functional.interpolate(
                torch.from_numpy(array), mode='bilinear',
                scale_factor=scale, 
            ).numpy()
        visdom.images(
            numpy.clip(array, 0.0, 1.0),
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        )

    @staticmethod #TODO: refactor these into a common module
    def __no_transform(tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    @staticmethod #TODO: refactor these into a common module
    def __minmax_normalization(tensor: torch.Tensor) -> torch.Tensor:
        b, _, __, ___ = tensor.size()
        min_v = torch.min(tensor.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        max_v = torch.max(tensor.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        return (tensor - min_v) / (max_v - min_v)

    @staticmethod #TODO: refactor these into a common module
    def __sum_minmax_norm(tensor: torch.Tensor) -> torch.Tensor:
        b, _, __, ___ = tensor.size()
        aggregated = torch.sum(tensor, dim=1, keepdim=True)
        min_v = torch.min(aggregated.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        max_v = torch.max(aggregated.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        return (aggregated - min_v) / (max_v - min_v)
            





import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import inspect
import typing
import itertools
import toolz
import logging
import functools

log = logging.getLogger(__name__)

#TODO: use class param execs

__all__ = ["Cascade"]

def _mul(
    tensor_dict: typing.Dict[str, torch.Tensor],
    lhs_key: str, rhs_key: str
) -> torch.Tensor:
    return tensor_dict[lhs_key] * tensor_dict[rhs_key]

def _add(
    tensor_dict: typing.Dict[str, torch.Tensor],
    keys: typing.Sequence[str],
) -> torch.Tensor:
    return tensor_dict[keys[0]] + tensor_dict[keys[1]]

def _sub(
    tensor_dict: typing.Dict[str, torch.Tensor],
    keys: typing.Sequence[str],
) -> torch.Tensor:
    return tensor_dict[keys[0]] - tensor_dict[keys[1]]

def _div(
    tensor_dict: typing.Dict[str, torch.Tensor],
    keys: typing.Sequence[str],
) -> torch.Tensor:
    return tensor_dict[keys[0]] / tensor_dict[keys[1]]

def _cat(
    tensor_dict: typing.Dict[str, torch.Tensor],
    keys: typing.Sequence[str],
) -> torch.Tensor:
    return torch.cat([tensor_dict[keys[0]], tensor_dict[keys[1]]], dim=1)

def _dict(
    tensor_dict: typing.Dict[str, torch.Tensor],
    keys: typing.Sequence[str],
) -> torch.Tensor:
    return toolz.get_in(keys, tensor_dict, no_default=True) #NOTE: should crash if the key is not found

__ACCESSORS__ = {
    ' * ': _mul,
    ' + ': _add,
    ' - ': _sub,
    ' / ': _div,
    ' | ': _cat,
    '.': _dict,
}

#TODO: need a lexer/parser/grammar here...
def _create_accessor(key: typing.Optional[typing.Union[str, typing.Sequence[str]]]):
    if key is None:
        return lambda _: None
    for k in __ACCESSORS__.keys():
        if k in key:
            return functools.partial(__ACCESSORS__[k], keys=key.split(k))
    return lambda td, k=key: td[k]


class Cascade(torch.nn.ModuleDict): #TODO: check if x: ['arg'] is the same as x: 'arg'
    execs: typing.List[typing.Callable]

    def __init__(self, 
        monads: omegaconf.DictConfig,
        **kwargs: typing.Mapping[str, typing.Any]
    ):
        super(Cascade, self).__init__()
        #TODO: first construct via monads and then create lambdas via kwargs
        loop = ((key, params) for key, params in kwargs.items() if key in monads.keys())
        #NOTE: check for not found keys and notify the potential error
        errors = [k for k in kwargs if k not in monads]
        if errors:
            log.error("Some monads were not found in the configuration and will be ignored!")
        self.execs = []
        for k, p in loop:
            #TODO: if instantiation fails, error out with message saying about adding it to the config file
            self.add_module(k, hyu.instantiate(getattr(monads, k))) #TODO: add construction loop first in class execs, and then funcs based on kwargs
            # last_module = toolz.last(self.modules()) #NOTE: moduledict is ordered            
            module = self[k]
            sig = inspect.signature(module.forward)
            props = [prop for prop in sig.parameters if p[prop] is not None]
            for keys in zip(*list(p[prop] for prop in itertools.chain(props, ['out']))):
                accessors = [_create_accessor(k if isinstance(k, str) or k is None else k[0]) for k in keys[:-1]]
                self.execs.append(lambda tensor_dict, 
                        acc=accessors, k=keys, p=props, f=module:
                    tensor_dict.update({
                        k[-1]: f(**dict(zip(p,                             
                            list(a(tensor_dict) if type(i) is str 
                                else list(tensor_dict[j] for j in i) if i is not None else None
                                for a, i in zip(acc, k[:-1])
                            )# list(tensor_dict[i] if type(i) is str
                        )))
                    })
                )                

    def forward(self,
        tensors: typing.Dict[str, torch.Tensor]    
    ) -> typing.Dict[str, torch.Tensor]:
        for exe in self.execs:
            exe(tensors) #NOTE: each executing lambda updates the tensor dict itself
        return tensors



import logging
import torch
import redner
import pyredner
import time
from typing import List, Union, Tuple, Optional
import warnings

log = logging.getLogger(__name__)

use_correlated_random_number = False
def set_use_correlated_random_number(v: bool):
    """
        | There is a bias-variance trade off in the backward pass.
        | If the forward pass and the backward pass are correlated
        | the gradients are biased for L2 loss.
        | E[d/dx(f(x) - y)^2] = E[(f(x) - y) d/dx f(x)]
        |                     = E[f(x) - y] E[d/dx f(x)]
        | The last equation only holds when f(x) and d/dx f(x) are independent.
        | It is usually better to use the unbiased one, but we left it as an option here
    """
    global use_correlated_random_number
    use_correlated_random_number = v

def get_use_correlated_random_number():
    """
        See set_use_correlated_random_number
    """
    global use_correlated_random_number
    return use_correlated_random_number

print_timing = False
def set_print_timing(v: bool):
    """
        Set whether to print time measurements or not.
    """
    global print_timing
    print_timing = v

def get_print_timing():
    """
        Get whether we print time measurements or not.
    """
    global print_timing
    return print_timing

def serialize_texture(texture, args, device):
    if texture is None:
        args.append(0)
        return
    args.append(len(texture.mipmap))
    for mipmap in texture.mipmap:
        assert(torch.isfinite(mipmap).all())
        assert(mipmap.is_contiguous())
        if mipmap.device != device:
            warnings.warn('Converting texture from {} to {}, this can be inefficient.'.format(mipmap.device, device))
        args.append(mipmap.to(device))
    assert(torch.isfinite(texture.uv_scale).all())
    args.append(texture.uv_scale.to(device))

class Context: pass

class RenderFunction(torch.autograd.Function):
    """
        The PyTorch interface of C++ redner.
    """

    @staticmethod
    def serialize_scene(scene: pyredner.Scene,
                        num_samples: Union[int, Tuple[int, int]],
                        max_bounces: int,
                        channels: List = [redner.channels.radiance],
                        sampler_type = redner.SamplerType.independent,
                        use_primary_edge_sampling: bool = True,
                        use_secondary_edge_sampling: bool = True,
                        sample_pixel_center: bool = False,
                        device: Optional[torch.device] = None):
        """
            Given a pyredner scene & rendering options, convert them to a linear list of argument,
            so that we can use it in PyTorch.

            Args
            ====
            scene: pyredner.Scene
            num_samples: int
                Number of samples per pixel for forward and backward passes.
                Can be an integer or a tuple of 2 integers.
                If a single integer is provided, use the same number of samples
                for both.
            max_bounces: int
                Number of bounces for global illumination,
                1 means direct lighting only.
            channels: List[redner.channels]
                | A list of channels that should present in the output image
                | following channels are supported\:
                | redner.channels.radiance,
                | redner.channels.alpha,
                | redner.channels.depth,
                | redner.channels.position,
                | redner.channels.geometry_normal,
                | redner.channels.shading_normal,
                | redner.channels.uv,
                | redner.channels.barycentric_coordinates,
                | redner.channels.diffuse_reflectance,
                | redner.channels.specular_reflectance,
                | redner.channels.vertex_color,
                | redner.channels.roughness,
                | redner.channels.generic_texture,
                | redner.channels.shape_id,
                | redner.channels.triangle_id,
                | redner.channels.material_id
                | all channels, except for shape id, triangle id, and material id, are differentiable
            sampler_type: redner.SamplerType
                | Which sampling pattern to use?
                | see `Chapter 7 of the PBRT book <http://www.pbr-book.org/3ed-2018/Sampling_and_Reconstruction.html>`
                  for an explanation of the difference between different samplers.
                | Following samplers are supported:
                | redner.SamplerType.independent
                | redner.SamplerType.sobol
            use_primary_edge_sampling: bool

            use_secondary_edge_sampling: bool

            sample_pixel_center: bool
                Always sample at the pixel center when rendering.
                This trades noise with aliasing.
                If this option is activated, the rendering becomes non-differentiable
                (since there is no antialiasing integral),
                and redner's edge sampling becomes an approximation to the gradients of the aliased rendering.

            device: Optional[torch.device]
                Which device should we store the data in.
                If set to None, use the device from pyredner.get_device().
        """
        if device is None:
            device = pyredner.get_device()
        if device.index is None and device.type == 'cuda':
            # Assign a default index so that we can avoid generating warnings
            device = torch.device('cuda:' + str(torch.cuda.current_device()))

        # Record if there is any parameter that requires gradient need discontinuity sampling.
        # For skipping edge sampling when it is not necessary.
        requires_visibility_grad = False
        cam = scene.camera
        num_shapes = len(scene.shapes)
        num_materials = len(scene.materials)
        num_lights = len(scene.area_lights)
        for light_id, light in enumerate(scene.area_lights):
            scene.shapes[light.shape_id].light_id = light_id

        if max_bounces == 0:
            use_secondary_edge_sampling = False

        args = []
        args.append(num_shapes)
        args.append(num_materials)
        args.append(num_lights)
        if not cam.position is None and not torch.isfinite(cam.position).all():
            log.warning("Camera position is not finite !")
        if not cam.look_at is None and not torch.isfinite(cam.look_at).all():
            log.warning("Camera look at is not finite !")
        if not cam.up is None and not torch.isfinite(cam.up).all():
            log.warning("Camera up is not finite !")        
        assert(torch.isfinite(cam.intrinsic_mat_inv).all())
        assert(torch.isfinite(cam.intrinsic_mat).all())
        if cam.position is not None and cam.position.requires_grad:
            requires_visibility_grad = True
        if cam.look_at is not None and cam.look_at.requires_grad:
            requires_visibility_grad = True
        if cam.up is not None and cam.up.requires_grad:
            requires_visibility_grad = True
        args.append(cam.position.cpu() if cam.position is not None else None)
        args.append(cam.look_at.cpu() if cam.look_at is not None else None)
        args.append(cam.up.cpu() if cam.up is not None else None)
        if cam.cam_to_world is not None:
            if cam.cam_to_world.requires_grad:
                requires_visibility_grad = True
            args.append(cam.cam_to_world.cpu().contiguous())
        else:
            args.append(None)
        if cam.world_to_cam is not None:
            if cam.world_to_cam.requires_grad:
                requires_visibility_grad = True
            args.append(cam.world_to_cam.cpu().contiguous())
        else:
            args.append(None)
        if cam.intrinsic_mat.requires_grad or cam.intrinsic_mat_inv.requires_grad:
            requires_visibility_grad = True
        args.append(cam.intrinsic_mat_inv.cpu().contiguous())
        args.append(cam.intrinsic_mat.cpu().contiguous())
        if cam.distortion_params is not None:
            if cam.distortion_params.requires_grad:
                requires_visibility_grad = True
            args.append(cam.distortion_params.cpu().contiguous())
        else:
            args.append(None)
        args.append(cam.clip_near)
        args.append(cam.resolution)
        viewport = cam.viewport
        if viewport is None:
            viewport = (0, 0, cam.resolution[0], cam.resolution[1])
        # Clamp the viewport if necessary
        viewport = (max(viewport[0], 0),
                    max(viewport[1], 0),
                    min(viewport[2], cam.resolution[0]),
                    min(viewport[3], cam.resolution[1]))
        args.append(viewport)
        args.append(cam.camera_type)
        for shape in scene.shapes:
            if not torch.isfinite(shape.vertices).all():
                log.warning("Rendering vertices which are not finite !")
            if (shape.uvs is not None):
                assert(torch.isfinite(shape.uvs).all())
            if (shape.normals is not None):
                assert(torch.isfinite(shape.normals).all())
            if (shape.vertices.requires_grad):
                requires_visibility_grad = True
            if shape.vertices.device != device:
                warnings.warn('Converting shape vertices from {} to {}, this can be inefficient.'.format(shape.vertices.device, device))
            if shape.indices.device != device:
                warnings.warn('Converting shape indices from {} to {}, this can be inefficient.'.format(shape.indices.device, device))
            args.append(shape.vertices.to(device))
            args.append(shape.indices.to(device))
            args.append(shape.uvs.to(device) if shape.uvs is not None else None)
            args.append(shape.normals.to(device) if shape.normals is not None else None)
            args.append(shape.uv_indices.to(device) if shape.uv_indices is not None else None)
            args.append(shape.normal_indices.to(device) if shape.normal_indices is not None else None)
            args.append(shape.colors.to(device) if shape.colors is not None else None)
            args.append(shape.material_id)
            args.append(shape.light_id)
        for material in scene.materials:
            serialize_texture(material.diffuse_reflectance, args, device)
            serialize_texture(material.specular_reflectance, args, device)
            serialize_texture(material.roughness, args, device)
            serialize_texture(material.generic_texture, args, device)
            serialize_texture(material.normal_map, args, device)
            args.append(material.compute_specular_lighting)
            args.append(material.two_sided)
            args.append(material.use_vertex_color)
        for light in scene.area_lights:
            args.append(light.shape_id)
            args.append(light.intensity.cpu())
            args.append(light.two_sided)
            args.append(light.directly_visible)
        if scene.envmap is not None:
            assert(torch.isfinite(scene.envmap.env_to_world).all())
            assert(torch.isfinite(scene.envmap.world_to_env).all())
            assert(torch.isfinite(scene.envmap.sample_cdf_ys).all())
            assert(torch.isfinite(scene.envmap.sample_cdf_xs).all())
            serialize_texture(scene.envmap.values, args, device)
            args.append(scene.envmap.env_to_world.cpu())
            args.append(scene.envmap.world_to_env.cpu())
            args.append(scene.envmap.sample_cdf_ys.to(device))
            args.append(scene.envmap.sample_cdf_xs.to(device))
            args.append(scene.envmap.pdf_norm)
            args.append(scene.envmap.directly_visible)
        else:
            args.append(None)

        args.append(num_samples)
        args.append(max_bounces)
        args.append(channels)
        args.append(sampler_type)
        if requires_visibility_grad:
            args.append(use_primary_edge_sampling)
            args.append(use_secondary_edge_sampling)
        else:
            # Don't need to do edge sampling if we don't require spatial derivatives
            args.append(False)
            args.append(False)
        args.append(sample_pixel_center)
        args.append(device)

        return args

    @staticmethod
    def unpack_args(seed,
                    args,
                    use_primary_edge_sampling = None,
                    use_secondary_edge_sampling = None):
        """
            Given a list of serialized scene arguments, unpack
            all information into a Context.
        """

        current_index = 0
        num_shapes = args[current_index]
        current_index += 1
        num_materials = args[current_index]
        current_index += 1
        num_lights = args[current_index]
        current_index += 1

        cam_position = args[current_index]
        current_index += 1
        cam_look_at = args[current_index]
        current_index += 1
        cam_up = args[current_index]
        current_index += 1
        cam_to_world = args[current_index]
        current_index += 1
        world_to_cam = args[current_index]
        current_index += 1
        intrinsic_mat_inv = args[current_index]
        current_index += 1
        intrinsic_mat = args[current_index]
        current_index += 1
        distortion_params = args[current_index]
        current_index += 1
        clip_near = args[current_index]
        current_index += 1
        resolution = args[current_index]
        current_index += 1
        viewport = args[current_index]
        current_index += 1
        camera_type = args[current_index]
        current_index += 1
        if cam_to_world is None:
            camera = redner.Camera(resolution[1],
                                   resolution[0],
                                   redner.float_ptr(cam_position.data_ptr()),
                                   redner.float_ptr(cam_look_at.data_ptr()),
                                   redner.float_ptr(cam_up.data_ptr()),
                                   redner.float_ptr(0), # cam_to_world
                                   redner.float_ptr(0), # world_to_cam
                                   redner.float_ptr(intrinsic_mat_inv.data_ptr()),
                                   redner.float_ptr(intrinsic_mat.data_ptr()),
                                   redner.float_ptr(distortion_params.data_ptr() if distortion_params is not None else 0),
                                   clip_near,
                                   camera_type,
                                   redner.Vector2i(viewport[1], viewport[0]),
                                   redner.Vector2i(viewport[3], viewport[2]))
        else:
            camera = redner.Camera(resolution[1],
                                   resolution[0],
                                   redner.float_ptr(0), # cam_position
                                   redner.float_ptr(0), # cam_look_at
                                   redner.float_ptr(0), # cam_up
                                   redner.float_ptr(cam_to_world.data_ptr()),
                                   redner.float_ptr(world_to_cam.data_ptr()),
                                   redner.float_ptr(intrinsic_mat_inv.data_ptr()),
                                   redner.float_ptr(intrinsic_mat.data_ptr()),
                                   redner.float_ptr(distortion_params.data_ptr() if distortion_params is not None else 0),
                                   clip_near,
                                   camera_type,
                                   redner.Vector2i(viewport[1], viewport[0]),
                                   redner.Vector2i(viewport[3], viewport[2]))
        shapes = []
        for i in range(num_shapes):
            vertices = args[current_index]
            current_index += 1
            indices = args[current_index]
            current_index += 1
            uvs = args[current_index]
            current_index += 1
            normals = args[current_index]
            current_index += 1
            uv_indices = args[current_index]
            current_index += 1
            normal_indices = args[current_index]
            current_index += 1
            colors = args[current_index]
            current_index += 1
            material_id = args[current_index]
            current_index += 1
            light_id = args[current_index]
            current_index += 1
            assert(vertices.is_contiguous())
            assert(indices.is_contiguous())
            if uvs is not None:
                assert(uvs.is_contiguous())
            if normals is not None:
                assert(normals.is_contiguous())
            if uv_indices is not None:
                assert(uv_indices.is_contiguous())
            if normal_indices is not None:
                assert(normal_indices.is_contiguous())
            shapes.append(redner.Shape(\
                redner.float_ptr(vertices.data_ptr()),
                redner.int_ptr(indices.data_ptr()),
                redner.float_ptr(uvs.data_ptr() if uvs is not None else 0),
                redner.float_ptr(normals.data_ptr() if normals is not None else 0),
                redner.int_ptr(uv_indices.data_ptr() if uv_indices is not None else 0),
                redner.int_ptr(normal_indices.data_ptr() if normal_indices is not None else 0),
                redner.float_ptr(colors.data_ptr() if colors is not None else 0),
                int(vertices.shape[0]),
                int(uvs.shape[0]) if uvs is not None else 0,
                int(normals.shape[0]) if normals is not None else 0,
                int(indices.shape[0]),
                material_id,
                light_id))

        materials = []
        for i in range(num_materials):
            num_levels = args[current_index]
            current_index += 1
            diffuse_reflectance = []
            for j in range(num_levels):
                diffuse_reflectance.append(args[current_index])
                current_index += 1
            diffuse_uv_scale = args[current_index]
            current_index += 1
            
            num_levels = args[current_index]
            current_index += 1
            specular_reflectance = []
            for j in range(num_levels):
                specular_reflectance.append(args[current_index])
                current_index += 1
            specular_uv_scale = args[current_index]
            current_index += 1
            
            num_levels = args[current_index]
            current_index += 1
            roughness = []
            for j in range(num_levels):
                roughness.append(args[current_index])
                current_index += 1
            roughness_uv_scale = args[current_index]
            current_index += 1

            num_levels = args[current_index]
            current_index += 1
            generic_texture = []
            if num_levels > 0:
                for j in range(num_levels):
                    generic_texture.append(args[current_index])
                    current_index += 1
                generic_uv_scale = args[current_index]
                current_index += 1
            else:
                generic_uv_scale = None

            num_levels = args[current_index]
            current_index += 1
            normal_map = []
            if num_levels > 0:
                for j in range(num_levels):
                    normal_map.append(args[current_index])
                    current_index += 1
                normal_map_uv_scale = args[current_index]
                current_index += 1
            else:
                normal_map_uv_scale = None

            compute_specular_lighting = args[current_index]
            current_index += 1
            two_sided = args[current_index]
            current_index += 1
            use_vertex_color = args[current_index]
            current_index += 1

            if diffuse_reflectance[0].dim() == 1:
                # Constant texture
                diffuse_reflectance = redner.Texture3(\
                    [redner.float_ptr(diffuse_reflectance[0].data_ptr())],
                    [0], [0], 3,
                    redner.float_ptr(diffuse_uv_scale.data_ptr()))
            else:
                assert(diffuse_reflectance[0].dim() == 3)
                diffuse_reflectance = redner.Texture3(\
                    [redner.float_ptr(x.data_ptr()) for x in diffuse_reflectance],
                    [x.shape[1] for x in diffuse_reflectance],
                    [x.shape[0] for x in diffuse_reflectance],
                    3,
                    redner.float_ptr(diffuse_uv_scale.data_ptr()))

            if specular_reflectance[0].dim() == 1:
                # Constant texture
                specular_reflectance = redner.Texture3(\
                    [redner.float_ptr(specular_reflectance[0].data_ptr())],
                    [0], [0], 3,
                    redner.float_ptr(specular_uv_scale.data_ptr()))
            else:
                assert(specular_reflectance[0].dim() == 3)
                specular_reflectance = redner.Texture3(\
                    [redner.float_ptr(x.data_ptr()) for x in specular_reflectance],
                    [x.shape[1] for x in specular_reflectance],
                    [x.shape[0] for x in specular_reflectance],
                    3,
                    redner.float_ptr(specular_uv_scale.data_ptr()))

            if roughness[0].dim() == 1:
                # Constant texture
                roughness = redner.Texture1(\
                    [redner.float_ptr(roughness[0].data_ptr())],
                    [0], [0], 1,
                    redner.float_ptr(roughness_uv_scale.data_ptr()))
            else:
                assert(roughness[0].dim() == 3)
                roughness = redner.Texture1(\
                    [redner.float_ptr(x.data_ptr()) for x in roughness],
                    [x.shape[1] for x in roughness],
                    [x.shape[0] for x in roughness],
                    1,
                    redner.float_ptr(roughness_uv_scale.data_ptr()))

            if len(generic_texture) > 0:
                assert(generic_texture[0].dim() == 3)
                generic_texture = redner.TextureN(\
                    [redner.float_ptr(x.data_ptr()) for x in generic_texture],
                    [x.shape[1] for x in generic_texture],
                    [x.shape[0] for x in generic_texture],
                    generic_texture[0].shape[2],
                    redner.float_ptr(generic_uv_scale.data_ptr()))
            else:
                generic_texture = redner.TextureN(\
                    [], [], [], 0, redner.float_ptr(0))

            if len(normal_map) > 0:
                assert(normal_map[0].dim() == 3)
                normal_map = redner.Texture3(\
                    [redner.float_ptr(x.data_ptr()) for x in normal_map],
                    [x.shape[1] for x in normal_map],
                    [x.shape[0] for x in normal_map],
                    3,
                    redner.float_ptr(normal_map_uv_scale.data_ptr()))
            else:
                normal_map = redner.Texture3(\
                    [], [], [], 0, redner.float_ptr(0))
            materials.append(redner.Material(\
                diffuse_reflectance,
                specular_reflectance,
                roughness,
                generic_texture,
                normal_map,
                compute_specular_lighting,
                two_sided,
                use_vertex_color))

        area_lights = []
        for i in range(num_lights):
            shape_id = args[current_index]
            current_index += 1
            intensity = args[current_index]
            current_index += 1
            two_sided = args[current_index]
            current_index += 1
            directly_visible = args[current_index]
            current_index += 1

            area_lights.append(redner.AreaLight(\
                shape_id,
                redner.float_ptr(intensity.data_ptr()),
                two_sided,
                directly_visible))

        envmap = None
        if args[current_index] is not None:
            num_levels = args[current_index]
            current_index += 1
            values = []
            for j in range(num_levels):
                values.append(args[current_index])
                current_index += 1
            envmap_uv_scale = args[current_index]
            current_index += 1
            env_to_world = args[current_index]
            current_index += 1
            world_to_env = args[current_index]
            current_index += 1
            sample_cdf_ys = args[current_index]
            current_index += 1
            sample_cdf_xs = args[current_index]
            current_index += 1
            pdf_norm = args[current_index]
            current_index += 1
            directly_visible = args[current_index]
            current_index += 1
            values = redner.Texture3(\
                [redner.float_ptr(x.data_ptr()) for x in values],
                [x.shape[1] for x in values], # width
                [x.shape[0] for x in values], # height
                3, # channels
                redner.float_ptr(envmap_uv_scale.data_ptr()))
            envmap = redner.EnvironmentMap(\
                values,
                redner.float_ptr(env_to_world.data_ptr()),
                redner.float_ptr(world_to_env.data_ptr()),
                redner.float_ptr(sample_cdf_ys.data_ptr()),
                redner.float_ptr(sample_cdf_xs.data_ptr()),
                pdf_norm,
                directly_visible)
        else:
            current_index += 1

        # Options
        num_samples = args[current_index]
        current_index += 1
        max_bounces = args[current_index]
        current_index += 1
        channels = args[current_index]
        current_index += 1
        sampler_type = args[current_index]
        current_index += 1
        use_primary_edge_sampling_ = args[current_index]
        current_index += 1
        use_secondary_edge_sampling_ = args[current_index]
        current_index += 1
        sample_pixel_center = args[current_index]
        current_index += 1
        device = args[current_index]
        current_index += 1

        if use_primary_edge_sampling is None:
            use_primary_edge_sampling = use_primary_edge_sampling_
        if use_secondary_edge_sampling is None:
            use_secondary_edge_sampling = use_secondary_edge_sampling_

        device_index = device.index
        if device.index is None:
            device_index = torch.cuda.current_device() if torch.cuda.is_available() else 0
        start = time.time()
        scene = redner.Scene(camera,
                             shapes,
                             materials,
                             area_lights,
                             envmap,
                             device.type == 'cuda',
                             device_index,
                             use_primary_edge_sampling,
                             use_secondary_edge_sampling)
        time_elapsed = time.time() - start
        if get_print_timing():
            print('Scene construction, time: %.5f s' % time_elapsed)

        # check that num_samples is a tuple
        if isinstance(num_samples, int):
            num_samples = (num_samples, num_samples)

        options = redner.RenderOptions(seed,
                                       num_samples[0],
                                       max_bounces,
                                       channels,
                                       sampler_type,
                                       sample_pixel_center)

        ctx = Context()
        ctx.channels = channels
        ctx.options = options
        ctx.resolution = resolution
        ctx.viewport = viewport
        ctx.scene = scene
        ctx.camera = camera
        ctx.shapes = shapes
        ctx.materials = materials
        ctx.area_lights = area_lights
        ctx.envmap = envmap
        ctx.scene = scene
        ctx.options = options
        ctx.num_samples = num_samples
        ctx.device = device

        return ctx

    @staticmethod
    def forward(ctx,
                seed,
                *args):
        """
            Forward rendering pass: given a serialized scene and output an image.
        """

        args_ctx = RenderFunction.unpack_args(seed, args)
        area_lights = args_ctx.area_lights
        camera = args_ctx.camera
        channels = args_ctx.channels
        envmap = args_ctx.envmap
        materials = args_ctx.materials
        num_samples = args_ctx.num_samples
        options = args_ctx.options
        resolution = args_ctx.resolution
        viewport = args_ctx.viewport
        scene = args_ctx.scene
        shapes = args_ctx.shapes
        device = args_ctx.device

        num_channels = redner.compute_num_channels(channels,
                                                   scene.max_generic_texture_dimension)
        img_height = viewport[2] - viewport[0]
        img_width = viewport[3] - viewport[1]
        rendered_image = torch.zeros(img_height, img_width, num_channels, device = device)
        start = time.time()
        redner.render(scene,
                      options,
                      redner.float_ptr(rendered_image.data_ptr()),
                      redner.float_ptr(0), # d_rendered_image
                      None, # d_scene
                      redner.float_ptr(0), # translational_gradient_image
                      redner.float_ptr(0)) # debug_image
        time_elapsed = time.time() - start
        if get_print_timing():
            print('Forward pass, time: %.5f s' % time_elapsed)

        ctx.camera = camera
        ctx.shapes = shapes
        ctx.materials = materials
        ctx.area_lights = area_lights
        ctx.envmap = envmap
        ctx.scene = scene
        ctx.options = options
        ctx.num_samples = num_samples
        ctx.device = device
        ctx.args = args # Important to prevent GC from deallocating the tensors
        return rendered_image

    @staticmethod
    def create_gradient_buffers(ctx):
        scene = ctx.scene
        options = ctx.options
        camera = ctx.camera
        device = ctx.device

        buffers = Context()

        if camera.use_look_at:
            buffers.d_cam_position = torch.zeros(3, device = device)
            buffers.d_cam_look = torch.zeros(3, device = device)
            buffers.d_cam_up = torch.zeros(3, device = device)
            buffers.d_cam_to_world = None
            buffers.d_world_to_cam = None
        else:
            buffers.d_cam_position = None
            buffers.d_cam_look = None
            buffers.d_cam_up = None
            buffers.d_cam_to_world = torch.zeros(4, 4, device = device)
            buffers.d_world_to_cam = torch.zeros(4, 4, device = device)
        buffers.d_intrinsic_mat_inv = torch.zeros(3, 3, device = device)
        buffers.d_intrinsic_mat = torch.zeros(3, 3, device = device)
        buffers.d_distortion_params = None
        if camera.has_distortion_params():
            buffers.d_distortion_params = torch.zeros(8, device = device)
        if camera.use_look_at:
            buffers.d_camera = redner.DCamera(\
                redner.float_ptr(buffers.d_cam_position.data_ptr()),
                redner.float_ptr(buffers.d_cam_look.data_ptr()),
                redner.float_ptr(buffers.d_cam_up.data_ptr()),
                redner.float_ptr(0), # cam_to_world
                redner.float_ptr(0), # world_to_cam
                redner.float_ptr(buffers.d_intrinsic_mat_inv.data_ptr()),
                redner.float_ptr(buffers.d_intrinsic_mat.data_ptr()),
                redner.float_ptr(buffers.d_distortion_params.data_ptr() if buffers.d_distortion_params is not None else 0))
        else:
            buffers.d_camera = redner.DCamera(\
                redner.float_ptr(0), # pos
                redner.float_ptr(0), # look
                redner.float_ptr(0), # up
                redner.float_ptr(buffers.d_cam_to_world.data_ptr()),
                redner.float_ptr(buffers.d_world_to_cam.data_ptr()),
                redner.float_ptr(buffers.d_intrinsic_mat_inv.data_ptr()),
                redner.float_ptr(buffers.d_intrinsic_mat.data_ptr()),
                redner.float_ptr(buffers.d_distortion_params.data_ptr() if buffers.d_distortion_params is not None else 0))
        buffers.d_vertices_list = []
        buffers.d_uvs_list = []
        buffers.d_normals_list = []
        buffers.d_colors_list = []
        buffers.d_shapes = []
        for shape in ctx.shapes:
            num_vertices = shape.num_vertices
            num_uv_vertices = shape.num_uv_vertices
            num_normal_vertices = shape.num_normal_vertices
            d_vertices = torch.zeros(num_vertices, 3, device = device)
            d_uvs = torch.zeros(num_uv_vertices, 2,
                device = device) if shape.has_uvs() else None
            d_normals = torch.zeros(num_normal_vertices, 3,
                device = device) if shape.has_normals() else None
            d_colors = torch.zeros(num_vertices, 3,
                device = device) if shape.has_colors() else None
            buffers.d_vertices_list.append(d_vertices)
            buffers.d_uvs_list.append(d_uvs)
            buffers.d_normals_list.append(d_normals)
            buffers.d_colors_list.append(d_colors)
            buffers.d_shapes.append(redner.DShape(\
                redner.float_ptr(d_vertices.data_ptr()),
                redner.float_ptr(d_uvs.data_ptr() if d_uvs is not None else 0),
                redner.float_ptr(d_normals.data_ptr() if d_normals is not None else 0),
                redner.float_ptr(d_colors.data_ptr() if d_colors is not None else 0)))

        buffers.d_diffuse_list = []
        buffers.d_diffuse_uv_scale_list = []
        buffers.d_specular_list = []
        buffers.d_specular_uv_scale_list = []
        buffers.d_roughness_list = []
        buffers.d_roughness_uv_scale_list = []
        buffers.d_generic_list = []
        buffers.d_generic_uv_scale_list = []
        buffers.d_normal_map_list = []
        buffers.d_normal_map_uv_scale_list = []
        buffers.d_materials = []
        for material in ctx.materials:
            if material.get_diffuse_size(0)[0] == 0:
                d_diffuse = [torch.zeros(3, device = device)]
            else:
                d_diffuse = []
                for l in range(material.get_diffuse_levels()):
                    diffuse_size = material.get_diffuse_size(l)
                    d_diffuse.append(\
                        torch.zeros(diffuse_size[1],
                                    diffuse_size[0],
                                    3, device = device))

            if material.get_specular_size(0)[0] == 0:
                d_specular = [torch.zeros(3, device = device)]
            else:
                d_specular = []
                for l in range(material.get_specular_levels()):
                    specular_size = material.get_specular_size(l)
                    d_specular.append(\
                        torch.zeros(specular_size[1],
                                    specular_size[0],
                                    3, device = device))

            if material.get_roughness_size(0)[0] == 0:
                d_roughness = [torch.zeros(1, device = device)]
            else:
                d_roughness = []
                for l in range(material.get_roughness_levels()):
                    roughness_size = material.get_roughness_size(l)
                    d_roughness.append(\
                        torch.zeros(roughness_size[1],
                                    roughness_size[0],
                                    1, device = device))

            if material.get_generic_levels() == 0:
                d_generic = None
            else:
                d_generic = []
                for l in range(material.get_generic_levels()):
                    generic_size = material.get_generic_size(l)
                    d_generic.append(\
                        torch.zeros(generic_size[2],
                                    generic_size[1],
                                    generic_size[0], device = device))

            if material.get_normal_map_levels() == 0:
                d_normal_map = None
            else:
                d_normal_map = []
                for l in range(material.get_normal_map_levels()):
                    normal_map_size = material.get_normal_map_size(l)
                    d_normal_map.append(\
                        torch.zeros(normal_map_size[1],
                                    normal_map_size[0],
                                    3, device = device))

            buffers.d_diffuse_list.append(d_diffuse)
            buffers.d_specular_list.append(d_specular)
            buffers.d_roughness_list.append(d_roughness)
            buffers.d_generic_list.append(d_generic)
            buffers.d_normal_map_list.append(d_normal_map)
            d_diffuse_uv_scale = torch.zeros(2, device = device)
            d_specular_uv_scale = torch.zeros(2, device = device)
            d_roughness_uv_scale = torch.zeros(2, device = device)
            buffers.d_diffuse_uv_scale_list.append(d_diffuse_uv_scale)
            buffers.d_specular_uv_scale_list.append(d_specular_uv_scale)
            buffers.d_roughness_uv_scale_list.append(d_roughness_uv_scale)
            if d_generic is None:
                d_generic_uv_scale = None
            else:
                d_generic_uv_scale = torch.zeros(2, device = device)
            if d_normal_map is None:
                d_normal_map_uv_scale = None
            else:
                d_normal_map_uv_scale = torch.zeros(2, device = device)

            buffers.d_generic_uv_scale_list.append(d_generic_uv_scale)
            buffers.d_normal_map_uv_scale_list.append(d_normal_map_uv_scale)
            if d_diffuse[0].dim() == 1:
                d_diffuse_tex = redner.Texture3(\
                    [redner.float_ptr(d_diffuse[0].data_ptr())],
                    [0],
                    [0],
                    3,
                    redner.float_ptr(d_diffuse_uv_scale.data_ptr()))
            else:
                d_diffuse_tex = redner.Texture3(\
                    [redner.float_ptr(x.data_ptr()) for x in d_diffuse],
                    [x.shape[1] for x in d_diffuse],
                    [x.shape[0] for x in d_diffuse],
                    3,
                    redner.float_ptr(d_diffuse_uv_scale.data_ptr()))

            if d_specular[0].dim() == 1:
                d_specular_tex = redner.Texture3(\
                    [redner.float_ptr(d_specular[0].data_ptr())],
                    [0],
                    [0],
                    3,
                    redner.float_ptr(d_specular_uv_scale.data_ptr()))
            else:
                d_specular_tex = redner.Texture3(\
                    [redner.float_ptr(x.data_ptr()) for x in d_specular],
                    [x.shape[1] for x in d_specular],
                    [x.shape[0] for x in d_specular],
                    3,
                    redner.float_ptr(d_specular_uv_scale.data_ptr()))

            if d_roughness[0].dim() == 1:
                d_roughness_tex = redner.Texture1(\
                    [redner.float_ptr(d_roughness[0].data_ptr())],
                    [0],
                    [0],
                    1,
                    redner.float_ptr(d_roughness_uv_scale.data_ptr()))
            else:
                d_roughness_tex = redner.Texture1(\
                    [redner.float_ptr(x.data_ptr()) for x in d_roughness],
                    [x.shape[1] for x in d_roughness],
                    [x.shape[0] for x in d_roughness],
                    1,
                    redner.float_ptr(d_roughness_uv_scale.data_ptr()))

            if d_generic is None:
                d_generic_tex = redner.TextureN(\
                    [], [], [], 0, redner.float_ptr(0))
            else:
                d_generic_tex = redner.TextureN(\
                    [redner.float_ptr(x.data_ptr()) for x in d_generic],
                    [x.shape[1] for x in d_generic],
                    [x.shape[0] for x in d_generic],
                    d_generic[0].shape[2],
                    redner.float_ptr(d_generic_uv_scale.data_ptr()))

            if d_normal_map is None:
                d_normal_map = redner.Texture3(\
                    [], [], [], 0, redner.float_ptr(0))
            else:
                d_normal_map = redner.Texture3(\
                    [redner.float_ptr(x.data_ptr()) for x in d_normal_map],
                    [x.shape[1] for x in d_normal_map],
                    [x.shape[0] for x in d_normal_map],
                    3,
                    redner.float_ptr(d_normal_map_uv_scale.data_ptr()))
            buffers.d_materials.append(redner.DMaterial(\
                d_diffuse_tex, d_specular_tex, d_roughness_tex,
                d_generic_tex, d_normal_map))

        buffers.d_intensity_list = []
        buffers.d_area_lights = []
        for light in ctx.area_lights:
            d_intensity = torch.zeros(3, device = device)
            buffers.d_intensity_list.append(d_intensity)
            buffers.d_area_lights.append(\
                redner.DAreaLight(redner.float_ptr(d_intensity.data_ptr())))

        buffers.d_envmap = None
        if ctx.envmap is not None:
            envmap = ctx.envmap
            buffers.d_envmap_values = []
            for l in range(envmap.get_levels()):
                size = envmap.get_size(l)
                buffers.d_envmap_values.append(\
                    torch.zeros(size[1],
                                size[0],
                                3, device = device))
            buffers.d_envmap_uv_scale = torch.zeros(2, device = device)
            d_envmap_tex = redner.Texture3(\
                [redner.float_ptr(x.data_ptr()) for x in buffers.d_envmap_values],
                [x.shape[1] for x in buffers.d_envmap_values],
                [x.shape[0] for x in buffers.d_envmap_values],
                3,
                redner.float_ptr(buffers.d_envmap_uv_scale.data_ptr()))
            buffers.d_world_to_env = torch.zeros(4, 4, device = device)
            buffers.d_envmap = redner.DEnvironmentMap(\
                d_envmap_tex,
                redner.float_ptr(buffers.d_world_to_env.data_ptr()))

        device_index = device.index
        if device.index is None:
            device_index = torch.cuda.current_device() if torch.cuda.is_available() else 0
        buffers.d_scene = redner.DScene(buffers.d_camera,
                                        buffers.d_shapes,
                                        buffers.d_materials,
                                        buffers.d_area_lights,
                                        buffers.d_envmap,
                                        device.type == 'cuda',
                                        device_index)
        return buffers

    @staticmethod
    def visualize_screen_gradient(grad_img: torch.Tensor,
                                  seed: int,
                                  scene: pyredner.Scene,
                                  num_samples: Union[int, Tuple[int, int]],
                                  max_bounces: int,
                                  channels: List = [redner.channels.radiance],
                                  sampler_type = redner.SamplerType.independent,
                                  use_primary_edge_sampling: bool = True,
                                  use_secondary_edge_sampling: bool = True,
                                  sample_pixel_center: bool = False):
        """
            Given a serialized scene and output an 2-channel image,
            which visualizes the derivatives of pixel color with respect to 
            the screen space coordinates.

            Args
            ====
            grad_img: Optional[torch.Tensor]
                The "adjoint" of the backpropagation gradient. If you don't know
                what this means just give None
            seed: int
                seed for the Monte Carlo random samplers
            See serialize_scene for the explanation of the rest of the arguments.
        """

        args = RenderFunction.serialize_scene(\
            scene = scene,
            num_samples = num_samples,
            max_bounces = max_bounces,
            sampler_type = sampler_type,
            channels = channels,
            sample_pixel_center = sample_pixel_center)
        args_ctx = RenderFunction.unpack_args(\
            seed, args, use_primary_edge_sampling, use_secondary_edge_sampling)
        channels = args_ctx.channels
        options = args_ctx.options
        resolution = args_ctx.resolution
        viewport = args_ctx.viewport
        scene = args_ctx.scene

        buffers = RenderFunction.create_gradient_buffers(args_ctx)
        num_channels = redner.compute_num_channels(channels,
                                                   scene.max_generic_texture_dimension)
        img_height = viewport[2] - viewport[0]
        img_width = viewport[3] - viewport[1]
        screen_gradient_image = torch.zeros(\
            img_height, img_width, 2, device = device)
        if grad_img is not None:
            assert(grad_img.shape[0] == img_height)
            assert(grad_img.shape[1] == img_width)
            assert(grad_img.shape[2] == num_channels)
        else:
            grad_img = torch.ones(img_height, img_width, num_channels, device = device)
        start = time.time()
        redner.render(scene,
                      options,
                      redner.float_ptr(0), # rendered_image
                      redner.float_ptr(grad_img.data_ptr()), # d_rendered_image
                      buffers.d_scene,
                      redner.float_ptr(screen_gradient_image.data_ptr()),
                      redner.float_ptr(0)) # debug_image
        time_elapsed = time.time() - start
        if get_print_timing():
            print('Visualize gradient, time: %.5f s' % time_elapsed)

        return screen_gradient_image

    @staticmethod
    def backward(ctx,
                 grad_img):
        if not grad_img.is_contiguous():
            grad_img = grad_img.contiguous()
        assert(torch.isfinite(grad_img).all())
        scene = ctx.scene
        options = ctx.options
        camera = ctx.camera

        buffers = RenderFunction.create_gradient_buffers(ctx)

        if not get_use_correlated_random_number():
            # Decouple the forward/backward random numbers by adding a big prime number
            options.seed += 1000003

        options.num_samples = ctx.num_samples[1]
        start = time.time()
        redner.render(scene, options,
                      redner.float_ptr(0), # rendered_image
                      redner.float_ptr(grad_img.data_ptr()),
                      buffers.d_scene,
                      redner.float_ptr(0), # translational_gradient_image
                      redner.float_ptr(0)) # debug_image
        time_elapsed = time.time() - start
        if get_print_timing():
            print('Backward pass, time: %.5f s' % time_elapsed)

        ret_list = []
        ret_list.append(None) # seed
        ret_list.append(None) # num_shapes
        ret_list.append(None) # num_materials
        ret_list.append(None) # num_lights
        if camera.use_look_at:
            ret_list.append(buffers.d_cam_position.cpu())
            ret_list.append(buffers.d_cam_look.cpu())
            ret_list.append(buffers.d_cam_up.cpu())
            ret_list.append(None) # cam_to_world
            ret_list.append(None) # world_to_cam
        else:
            ret_list.append(None) # pos
            ret_list.append(None) # look
            ret_list.append(None) # up
            ret_list.append(buffers.d_cam_to_world.cpu())
            ret_list.append(buffers.d_world_to_cam.cpu())
        ret_list.append(buffers.d_intrinsic_mat_inv.cpu())
        ret_list.append(buffers.d_intrinsic_mat.cpu())
        if not camera.has_distortion_params():
            ret_list.append(None) # distortion_params
        else:
            ret_list.append(buffers.d_distortion_params.cpu())
        ret_list.append(None) # clip near
        ret_list.append(None) # resolution
        ret_list.append(None) # viewport
        ret_list.append(None) # camera_type

        num_shapes = len(ctx.shapes)
        for i in range(num_shapes):
            ret_list.append(buffers.d_vertices_list[i])
            ret_list.append(None) # indices
            ret_list.append(buffers.d_uvs_list[i])
            ret_list.append(buffers.d_normals_list[i])
            ret_list.append(None) # uv_indices
            ret_list.append(None) # normal_indices
            ret_list.append(buffers.d_colors_list[i])
            ret_list.append(None) # material id
            ret_list.append(None) # light id

        num_materials = len(ctx.materials)
        for i in range(num_materials):
            ret_list.append(None) # num_levels
            for d_diffuse in buffers.d_diffuse_list[i]:
                ret_list.append(d_diffuse)
            ret_list.append(buffers.d_diffuse_uv_scale_list[i])
            ret_list.append(None) # num_levels
            for d_specular in buffers.d_specular_list[i]:
                ret_list.append(d_specular)
            ret_list.append(buffers.d_specular_uv_scale_list[i])
            ret_list.append(None) # num_levels
            for d_roughness in buffers.d_roughness_list[i]:
                ret_list.append(d_roughness)
            ret_list.append(buffers.d_roughness_uv_scale_list[i])
            if buffers.d_generic_list[i] is None:
                ret_list.append(None) # num_levels
            else:
                ret_list.append(None) # num_levels
                for d_generic in buffers.d_generic_list[i]:
                    ret_list.append(d_generic)
                ret_list.append(buffers.d_generic_uv_scale_list[i])
            if buffers.d_normal_map_list[i] is None:
                ret_list.append(None) # num_levels
            else:
                ret_list.append(None) # num_levels
                for d_normal_map in buffers.d_normal_map_list[i]:
                    ret_list.append(d_normal_map)
                ret_list.append(buffers.d_normal_map_uv_scale_list[i])
            ret_list.append(None) # compute_specular_lighting
            ret_list.append(None) # two sided
            ret_list.append(None) # use_vertex_color

        num_area_lights = len(ctx.area_lights)
        for i in range(num_area_lights):
            ret_list.append(None) # shape id
            ret_list.append(buffers.d_intensity_list[i].cpu())
            ret_list.append(None) # two_sided
            ret_list.append(None) # directly_visible

        if ctx.envmap is not None:
            ret_list.append(None) # num_levels
            for d_values in buffers.d_envmap_values:
                ret_list.append(d_values)
            ret_list.append(buffers.d_envmap_uv_scale)
            ret_list.append(None) # env_to_world
            ret_list.append(buffers.d_world_to_env.cpu())
            ret_list.append(None) # sample_cdf_ys
            ret_list.append(None) # sample_cdf_xs
            ret_list.append(None) # pdf_norm
            ret_list.append(None) # directly_visible
        else:
            ret_list.append(None)

        ret_list.append(None) # num samples
        ret_list.append(None) # num bounces
        ret_list.append(None) # channels
        ret_list.append(None) # sampler type
        ret_list.append(None) # use_primary_edge_sampling
        ret_list.append(None) # use_secondary_edge_sampling
        ret_list.append(None) # sample_pixel_center
        ret_list.append(None) # device

        return tuple(ret_list)


import pyredner
import random
import redner
import torch
import math
from typing import Union, Tuple, Optional, List

class DeferredLight:
    pass

class AmbientLight(DeferredLight):
    """
        Ambient light for deferred rendering.
    """
    def __init__(self,
                 intensity: torch.Tensor):
        self.intensity = intensity

    def render(self,
               position: torch.Tensor,
               normal: torch.Tensor,
               albedo: torch.Tensor):
        return self.intensity.to(albedo.device) * albedo

class PointLight(DeferredLight):
    """
        Point light with squared distance falloff for deferred rendering.
    """
    def __init__(self,
                 position: torch.Tensor,
                 intensity: torch.Tensor):
        self.position = position
        self.intensity = intensity

    def render(self,
               position: torch.Tensor,
               normal: torch.Tensor,
               albedo: torch.Tensor):
        light_dir = self.position.to(position.device) - position
        # the d^2 term:
        light_dist_sq = torch.sum(light_dir * light_dir, dim = -1, keepdim = True)
        light_dist = torch.sqrt(light_dist_sq)
        # Normalize light direction
        light_dir = light_dir / light_dist
        dot_l_n = torch.sum(light_dir * normal, dim = -1, keepdim = True)
        dot_l_n = torch.max(dot_l_n, torch.zeros_like(dot_l_n))
        return self.intensity.to(dot_l_n.device) * dot_l_n * (albedo / math.pi) / light_dist_sq 

class DirectionalLight(DeferredLight):
    """
        Directional light for deferred rendering.
    """
    def __init__(self,
                 direction: torch.Tensor,
                 intensity: torch.Tensor):
        self.direction = direction
        self.intensity = intensity

    def render(self,
               position: torch.Tensor,
               normal: torch.Tensor,
               albedo: torch.Tensor):
        # Normalize light direction
        light_dir = -self.direction / torch.norm(self.direction)
        light_dir = light_dir.view(1, 1, 3)
        light_dir = light_dir.to(normal.device)
        dot_l_n = torch.sum(light_dir * normal, dim = -1, keepdim = True)
        dot_l_n = torch.max(dot_l_n, torch.zeros_like(dot_l_n))
        return self.intensity.to(dot_l_n.device) * dot_l_n * (albedo / math.pi)

class SpotLight(DeferredLight):
    """
        Spot light with cosine falloff for deferred rendering.
        Note that we do not provide the cosine cutoff here since it is not
        differentiable.
    """
    def __init__(self,
                 position: torch.Tensor,
                 spot_direction: torch.Tensor,
                 spot_exponent: torch.Tensor,
                 intensity: torch.Tensor):
        self.position = position
        self.spot_direction = spot_direction
        self.spot_exponent = spot_exponent
        self.intensity = intensity

    def render(self,
               position: torch.Tensor,
               normal: torch.Tensor,
               albedo: torch.Tensor):
        light_dir = self.position.to(position.device) - position
        # Normalize light direction
        light_dir = light_dir / torch.norm(light_dir, dim = -1, keepdim = True)
        # Normalize spot direction
        spot_direction = -self.spot_direction / torch.norm(self.spot_direction)
        spot_direction = spot_direction.to(light_dir.device)
        spot_cosine = torch.sum(light_dir * spot_direction, dim = -1, keepdim = True)
        spot_cosine = torch.max(spot_cosine, torch.zeros_like(spot_cosine))
        spot_factor = torch.pow(spot_cosine, self.spot_exponent.to(spot_cosine.device))
        dot_l_n = torch.sum(light_dir * normal, dim = -1, keepdim = True)
        dot_l_n = torch.max(dot_l_n, torch.zeros_like(dot_l_n))
        return self.intensity.to(spot_factor.device) * spot_factor * dot_l_n * (albedo / math.pi)

def render_deferred(scene: Union[pyredner.Scene, List[pyredner.Scene]],
                    lights: Union[List[DeferredLight], List[List[DeferredLight]]],
                    alpha: bool = False,
                    aa_samples: int = 2,
                    seed: Optional[Union[int, List[int]]] = None,
                    sample_pixel_center: bool = False,
                    device: Optional[torch.device] = None):
    """
        Render the scenes using `deferred rendering <https://en.wikipedia.org/wiki/Deferred_shading>`_.
        We generate G-buffer images containing world-space position,
        normal, and albedo using redner, then shade the G-buffer
        using PyTorch code. Assuming Lambertian shading and does not
        compute shadow.

        Args
        ====
        scene: Union[pyredner.Scene, List[pyredner.Scene]]
            pyredner Scene containing camera, geometry and material.
            Can be a single scene or a list for batch render.
            For batch rendering all scenes need to have the same resolution.
        lights: Union[List[DeferredLight], List[List[DeferredLight]]]
            Lights for deferred rendering. If the scene is a list, and only
            a single list of lights is provided, the same lights are applied
            to all scenes. If a list of lists of lights is provided, each scene
            is lit by the corresponding lights.
        alpha: bool
            If set to False, generates a 3-channel image,
            otherwise generates a 4-channel image where the
            fourth channel is alpha.
        aa_samples: int
            Number of samples used for anti-aliasing at both x, y dimensions
            (e.g. if aa_samples=2, 4 samples are used).
        seed: Optional[Union[int, List[int]]]
            Random seed used for sampling. Randomly assigned if set to None.
            For batch render, if seed it not None, need to provide a list
            of seeds.
        sample_pixel_center: bool
            Always sample at the pixel center when rendering.
            This trades noise with aliasing.
            If this option is activated, the rendering becomes non-differentiable
            (since there is no antialiasing integral),
            and redner's edge sampling becomes an approximation to the gradients of the aliased rendering.
        device: Optional[torch.device]
            Which device should we store the data in.
            If set to None, use the device from pyredner.get_device().

        Returns
        =======
        torch.Tensor or List[torch.Tensor]
            | if input scene is a list: a tensor with size [N, H, W, C], N is the list size
            | else: a tensor with size [H, W, C]
            | if alpha == True, C = 4.
            | else, C = 3.
    """
    if device is None:
        device = pyredner.get_device()

    channels = [redner.channels.position,
                redner.channels.shading_normal,
                redner.channels.diffuse_reflectance]
    if alpha:
        channels.append(redner.channels.alpha)
    if isinstance(scene, pyredner.Scene):
        if seed == None:
            seed = random.randint(0, 16777216)
        # We do full-screen anti-aliasing: increase the rendering resolution
        # and downsample it after lighting
        org_res = scene.camera.resolution
        org_viewport = scene.camera.viewport
        scene.camera.resolution = (org_res[0] * aa_samples,
                                   org_res[1] * aa_samples)
        if org_viewport is not None:
            scene.camera.viewport = [i * aa_samples for i in org_viewport]
        scene_args = RenderFunction.serialize_scene(\
            scene = scene,
            num_samples = (1, 1),
            max_bounces = 0,
            sampler_type = redner.SamplerType.sobol,
            channels = channels,
            use_secondary_edge_sampling = False,
            sample_pixel_center = sample_pixel_center,
            device = device)
        # Need to revert the resolution back
        scene.camera.resolution = org_res
        scene.camera.viewport = org_viewport
        g_buffer = RenderFunction.apply(seed, *scene_args)
        pos = g_buffer[:, :, :3]
        normal = g_buffer[:, :, 3:6]
        albedo = g_buffer[:, :, 6:9]
        img = torch.zeros(g_buffer.shape[0], g_buffer.shape[1], 3, device = device)
        for light in lights:
            img = img + light.render(pos, normal, albedo)
        if alpha:
            # alpha is in the last channel
            img = torch.cat((img, g_buffer[:, :, 9:10]), dim = -1)
        if aa_samples > 1:
            # Downsample
            img = img.permute(2, 0, 1) # HWC -> CHW
            img = img.unsqueeze(0) # CHW -> NCHW
            if org_viewport is not None:
                org_size = org_viewport[2] - org_viewport[0], org_viewport[3] - org_viewport[1]
            else:
                org_size = org_res
            img = torch.nn.functional.interpolate(img, size = org_size, mode = 'area')
            img = img.squeeze(dim = 0) # NCHW -> CHW
            img = img.permute(1, 2, 0)
        return img
    else:
        assert(isinstance(scene, list))
        if seed == None:
            # Randomly generate a list of seed
            seed = []
            for i in range(len(scene)):
                seed.append(random.randint(0, 16777216))
        assert(len(seed) == len(scene))
        if len(lights) > 0 and not isinstance(lights[0], list):
            # Specialize version: stack g buffers and light all images in parallel
            g_buffers = []
            # Render each scene in the batch and stack them together
            for sc, se in zip(scene, seed):
                # We do full-screen anti-aliasing: increase the rendering resolution
                # and downsample it after lighting
                org_res = sc.camera.resolution
                org_viewport = sc.camera.viewport
                sc.camera.resolution = (org_res[0] * aa_samples,
                                        org_res[1] * aa_samples)
                if org_viewport is not None:
                    sc.camera.viewport = [i * aa_samples for i in org_viewport]
                scene_args = RenderFunction.serialize_scene(\
                    scene = sc,
                    num_samples = (1, 1),
                    max_bounces = 0,
                    sampler_type = redner.SamplerType.sobol,
                    channels = channels,
                    use_secondary_edge_sampling = False,
                    sample_pixel_center = sample_pixel_center,
                    device = device)
                # Need to revert the resolution back
                sc.camera.resolution = org_res
                sc.camera.viewport = org_viewport
                g_buffers.append(RenderFunction.apply(se, *scene_args))
            g_buffers = torch.stack(g_buffers)
            pos = g_buffers[:, :, :, :3]
            normal = g_buffers[:, :, :, 3:6]
            albedo = g_buffers[:, :, :, 6:9]
            imgs = torch.zeros(g_buffers.shape[0],
                               g_buffers.shape[1],
                               g_buffers.shape[2],
                               3,
                               device = device)
            for light in lights:
                imgs = imgs + light.render(pos, normal, albedo)
            if alpha:
                imgs = torch.cat((imgs, g_buffers[:, :, :, 9:10]), dim = -1)
        else:
            # If each scene has a different lighting: light them in the loop
            imgs = []
            # Render each scene in the batch and stack them together
            for sc, se, lgts in zip(scene, seed, lights):
                # We do full-screen anti-aliasing: increase the rendering resolution
                # and downsample it after lighting
                org_res = sc.camera.resolution
                org_viewport = sc.camera.viewport
                sc.camera.resolution = (org_res[0] * aa_samples,
                                        org_res[1] * aa_samples)
                if org_viewport is not None:
                    sc.camera.viewport = [i * aa_samples for i in org_viewport]
                scene_args = RenderFunction.serialize_scene(\
                    scene = sc,
                    num_samples = (1, 1),
                    max_bounces = 0,
                    sampler_type = redner.SamplerType.sobol,
                    channels = channels,
                    use_secondary_edge_sampling = False,
                    sample_pixel_center = sample_pixel_center,
                    device = device)
                # Need to revert the resolution back
                sc.camera.resolution = org_res
                sc.camera.viewport = org_viewport
                g_buffer = RenderFunction.apply(se, *scene_args)
                pos = g_buffer[:, :, :3]
                normal = g_buffer[:, :, 3:6]
                albedo = g_buffer[:, :, 6:9]
                img = torch.zeros(g_buffer.shape[0],
                                  g_buffer.shape[1],
                                  3,
                                  device = device)
                for light in lgts:
                    img = img + light.render(pos, normal, albedo)
                if alpha:
                    # alpha is in the last channel
                    img = torch.cat((img, g_buffer[:, :, 9:10]), dim = -1)
                imgs.append(img)
            imgs = torch.stack(imgs)
        if aa_samples > 1:
            # Downsample
            imgs = imgs.permute(0, 3, 1, 2) # NHWC -> NCHW
            if org_viewport is not None:
                org_size = org_viewport[2] - org_viewport[0], org_viewport[3] - org_viewport[1]
            else:
                org_size = org_res
            imgs = torch.nn.functional.interpolate(imgs, size = org_size, mode = 'area')
            imgs = imgs.permute(0, 2, 3, 1) # NCHW -> NHWC
        return imgs

def render_generic(scene: pyredner.Scene,
                   channels: List,
                   max_bounces: int = 1,
                   sampler_type = pyredner.sampler_type.sobol,
                   num_samples: Union[int, Tuple[int, int]] = (4, 4),
                   seed: Optional[int] = None,
                   sample_pixel_center: bool = False,
                   device: Optional[torch.device] = None):
    """
        A generic rendering function that can be either pathtracing or
        g-buffer rendering or both.

        Args
        ====
        scene: Union[pyredner.Scene, List[pyredner.Scene]]
            pyredner Scene containing camera, geometry and material.
            Can be a single scene or a list for batch render.
            For batch rendering all scenes need to have the same resolution.
        channels: List[pyredner.channels]
            | A list of the following channels\:
            | pyredner.channels.alpha
            | pyredner.channels.depth
            | pyredner.channels.position
            | pyredner.channels.geometry_normal
            | pyredner.channels.shading_normal
            | pyredner.channels.uv
            | pyredner.channels.barycentric_coordinates
            | pyredner.channels.diffuse_reflectance
            | pyredner.channels.specular_reflectance
            | pyredner.channels.roughness
            | pyredner.channels.generic_texture
            | pyredner.channels.vertex_color
            | pyredner.channels.shape_id
            | pyredner.channels.triangle_id
            | pyredner.channels.material_id
        max_bounces: int
            Number of bounces for global illumination, 1 means direct lighting only.
        sampler_type: pyredner.sampler_type
            | Which sampling pattern to use? See 
              `Chapter 7 of the PBRT book <http://www.pbr-book.org/3ed-2018/Sampling_and_Reconstruction.html>`_
              for an explanation of the difference between different samplers.
            | Following samplers are supported\:
            | pyredner.sampler_type.independent
            | pyredner.sampler_type.sobol
        num_samples: int
            Number of samples per pixel for forward and backward passes.
            Can be an integer or a tuple of 2 integers.
        seed: Optional[Union[int, List[int]]]
            Random seed used for sampling. Randomly assigned if set to None.
            For batch render, if seed it not None, need to provide a list
            of seeds.
        sample_pixel_center: bool
            Always sample at the pixel center when rendering.
            This trades noise with aliasing.
            If this option is activated, the rendering becomes non-differentiable
            (since there is no antialiasing integral),
            and redner's edge sampling becomes an approximation to the gradients of the aliased rendering.
        device: Optional[torch.device]
            Which device should we store the data in.
            If set to None, use the device from pyredner.get_device().

        Returns
        =======
        torch.Tensor or List[torch.Tensor]
            | if input scene is a list: a tensor with size [N, H, W, C], N is the list size
            | else: a tensor with size [H, W, C]
    """
    if device is None:
        device = pyredner.get_device()

    if isinstance(scene, pyredner.Scene):
        if seed==None:
            seed = random.randint(0, 16777216)
        scene_args = RenderFunction.serialize_scene(\
            scene = scene,
            num_samples = num_samples,
            max_bounces = max_bounces,
            sampler_type = sampler_type,
            channels = channels,
            sample_pixel_center = sample_pixel_center,
            device = device)
        return RenderFunction.apply(seed, *scene_args)
    else:
        assert(isinstance(scene, list))
        if seed == None:
            # Randomly generate a list of seed
            seed = []
            for i in range(len(scene)):
                seed.append(random.randint(0, 16777216))
        assert(len(seed) == len(scene))
        # Render each scene in the batch and stack them together
        imgs = []
        for sc, se in zip(scene, seed):
            scene_args = RenderFunction.serialize_scene(\
                scene = sc,
                num_samples = num_samples,
                max_bounces = max_bounces,
                sampler_type = sampler_type,
                channels = channels,
                sample_pixel_center = sample_pixel_center,
                device = device)
            imgs.append(RenderFunction.apply(se, *scene_args))
        imgs = torch.stack(imgs)
        return imgs

def render_g_buffer(scene: Union[pyredner.Scene, List[pyredner.Scene]],
                    channels: List,
                    num_samples: Union[int, Tuple[int, int]] = (1, 1),
                    seed: Optional[Union[int, List[int]]] = None,
                    sample_pixel_center: bool = False,
                    device: Optional[torch.device] = None):
    """
        Render G buffers from the scene.

        Args
        ====
        scene: Union[pyredner.Scene, List[pyredner.Scene]]
            pyredner Scene containing camera, geometry and material.
            Can be a single scene or a list for batch render.
            For batch rendering all scenes need to have the same resolution.
        channels: List[pyredner.channels]
            | A list of the following channels\:
            | pyredner.channels.alpha
            | pyredner.channels.depth
            | pyredner.channels.position
            | pyredner.channels.geometry_normal
            | pyredner.channels.shading_normal
            | pyredner.channels.uv
            | pyredner.channels.barycentric_coordinates
            | pyredner.channels.diffuse_reflectance
            | pyredner.channels.specular_reflectance
            | pyredner.channels.roughness
            | pyredner.channels.generic_texture
            | pyredner.channels.vertex_color
            | pyredner.channels.shape_id
            | pyredner.channels.triangle_id
            | pyredner.channels.material_id
        num_samples: Union[int, Tuple[int, int]]
            Number of samples for forward and backward passes, respectively.
            If a single integer is provided, use the same number of samples
            for both.
        seed: Optional[Union[int, List[int]]]
            Random seed used for sampling. Randomly assigned if set to None.
            For batch render, if seed it not None, need to provide a list
            of seeds.
        sample_pixel_center: bool
            Always sample at the pixel center when rendering.
            This trades noise with aliasing.
            If this option is activated, the rendering becomes non-differentiable
            (since there is no antialiasing integral),
            and redner's edge sampling becomes an approximation to the gradients of the aliased rendering.
        device: Optional[torch.device]
            Which device should we store the data in.
            If set to None, use the device from pyredner.get_device().

        Returns
        =======
        torch.Tensor or List[torch.Tensor]
            | if input scene is a list: a tensor with size [N, H, W, C], N is the list size
            | else: a tensor with size [H, W, C]
    """
    return render_generic(scene = scene,
                          channels = channels,
                          max_bounces = 0,
                          sampler_type = redner.SamplerType.sobol,
                          num_samples = num_samples,
                          seed = seed,
                          sample_pixel_center = sample_pixel_center,
                          device = device)

def render_pathtracing(scene: Union[pyredner.Scene, List[pyredner.Scene]],
                       alpha: bool = False,
                       max_bounces: int = 1,
                       sampler_type = pyredner.sampler_type.sobol,
                       num_samples: Union[int, Tuple[int, int]] = (4, 4),
                       seed: Optional[Union[int, List[int]]] = None,
                       sample_pixel_center: bool = False,
                       device: Optional[torch.device] = None):
    """
        Render a pyredner scene using pathtracing.

        Args
        ====
        scene: Union[pyredner.Scene, List[pyredner.Scene]]
            pyredner Scene containing camera, geometry and material.
            Can be a single scene or a list for batch render.
            For batch rendering all scenes need to have the same resolution.
        max_bounces: int
            Number of bounces for global illumination, 1 means direct lighting only.
        sampler_type: pyredner.sampler_type
            | Which sampling pattern to use? See 
              `Chapter 7 of the PBRT book <http://www.pbr-book.org/3ed-2018/Sampling_and_Reconstruction.html>`_
              for an explanation of the difference between different samplers.
            | Following samplers are supported\:
            | pyredner.sampler_type.independent
            | pyredner.sampler_type.sobol
        num_samples: int
            Number of samples per pixel for forward and backward passes.
            Can be an integer or a tuple of 2 integers.
        seed: Optional[Union[int, List[int]]]
            Random seed used for sampling. Randomly assigned if set to None.
            For batch render, if seed it not None, need to provide a list
            of seeds.
        sample_pixel_center: bool
            Always sample at the pixel center when rendering.
            This trades noise with aliasing.
            If this option is activated, the rendering becomes non-differentiable
            (since there is no antialiasing integral),
            and redner's edge sampling becomes an approximation to the gradients of the aliased rendering.
        device: Optional[torch.device]
            Which device should we store the data in.
            If set to None, use the device from pyredner.get_device().

        Returns
        =======
        torch.Tensor or List[torch.Tensor]
            | if input scene is a list: a tensor with size [N, H, W, C], N is the list size
            | else: a tensor with size [H, W, C]
            | if alpha == True, C = 4.
            | else, C = 3.
    """
    channels = [redner.channels.radiance]
    if alpha:
        channels.append(redner.channels.alpha)
    return render_generic(scene = scene,
                          channels = channels,
                          max_bounces = max_bounces,
                          sampler_type = sampler_type,
                          num_samples = num_samples,
                          seed = seed,
                          sample_pixel_center = sample_pixel_center,
                          device = device)

def render_albedo(scene: Union[pyredner.Scene, List[pyredner.Scene]],
                  alpha: bool = False,
                  num_samples: Union[int, Tuple[int, int]] = (16, 4),
                  seed: Optional[Union[int, List[int]]] = None,
                  sample_pixel_center: bool = False,
                  device: Optional[torch.device] = None):
    """
        Render the diffuse albedo colors of the scenes.

        Args
        ====
        scene: Union[pyredner.Scene, List[pyredner.Scene]]
            pyredner Scene containing camera, geometry and material.
            Can be a single scene or a list for batch render.
            For batch rendering all scenes need to have the same resolution.
        alpha: bool
            If set to False, generates a 3-channel image,
            otherwise generates a 4-channel image where the
            fourth channel is alpha.
        num_samples: Union[int, Tuple[int, int]]
            number of samples for forward and backward passes, respectively
            if a single integer is provided, use the same number of samples
            for both
        seed: Optional[Union[int, List[int]]]
            Random seed used for sampling. Randomly assigned if set to None.
            For batch render, if seed it not None, need to provide a list
            of seeds.
        sample_pixel_center: bool
            Always sample at the pixel center when rendering.
            This trades noise with aliasing.
            If this option is activated, the rendering becomes non-differentiable
            (since there is no antialiasing integral),
            and redner's edge sampling becomes an approximation to the gradients of the aliased rendering.
        device: Optional[torch.device]
            Which device should we store the data in.
            If set to None, use the device from pyredner.get_device().

        Returns
        =======
        torch.Tensor or List[torch.Tensor]
            | if input scene is a list: a tensor with size [N, H, W, C], N is the list size
            | else: a tensor with size [H, W, C]
            | if alpha == True, C = 4.
            | else, C = 3.
    """
    channels = [redner.channels.diffuse_reflectance]
    if alpha:
        channels.append(redner.channels.alpha)
    return render_g_buffer(scene = scene,
                           channels = channels,
                           num_samples = num_samples,
                           seed = seed,
                           sample_pixel_center = sample_pixel_center,
                           device = device)



import math
import torch
import typing

__all__ = ['AdaFactor']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/adafactor.py

class AdaFactor(torch.optim.Optimizer):
    """Implements AdaFactor algorithm.

    - **Paper**: [AdaFactor: Adaptive Learning Rates with Sublinear Memory Cost](https://arxiv.org/pdf/1804.04235.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: external learning rate (default: None)
        eps2: regularization constans for square gradient
            and parameter scale respectively (default: (1e-30, 1e-3))
        clip_threshold: threshold of root mean square of
            final gradient update (default: 1.0)
        decay_rate: coefficient used to compute running averages of square
            gradient (default: -0.8)
        beta1: coefficient used for computing running averages of gradient
            (default: None)
        weight_decay: weight decay (L2 penalty) (default: 0)
        scale_parameter: if true, learning rate is scaled by root mean square
            of parameter (default: True)
        relative_step: if true, time-dependent learning rate is computed
            instead of external learning rate (default: True)
        warmup_init: time-dependent learning rate computation depends on
            whether warm-up initialization is being used (default: False)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.AdaFactor(model.parameters())
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ https://arxiv.org/abs/1804.04235

    Note:
        Reference code: https://github.com/pytorch/fairseq/blob/master/fairseq/optim/adafactor.py  # noqa
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: typing.Optional[float]=None,
        eps2: typing.Tuple[float, float]=(1e-30, 1e-3),
        clip_threshold: float=1.0,
        decay_rate: float=-0.8,
        beta1: typing.Optional[float]=None,
        weight_decay: float=0.0,
        scale_parameter: bool=True,
        relative_step: bool=True,
        warmup_init: bool=False,
    ):
        if lr is not None and lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))
        if weight_decay < 0.0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )

        defaults = dict(
            lr=lr,
            eps2=eps2,
            clip_threshold=clip_threshold,
            decay_rate=decay_rate,
            beta1=beta1,
            weight_decay=weight_decay,
            scale_parameter=scale_parameter,
            relative_step=relative_step,
            warmup_init=warmup_init,
        )
        super(AdaFactor, self).__init__(params, defaults)

    def _get_lr(self, param_group: typing.Dict[str, typing.Any], param_state: typing.Dict[str, typing.Any]) -> float:
        rel_step_sz = param_group['lr']
        if param_group['relative_step']:
            min_step = (
                1e-6 * param_state['step']
                if param_group['warmup_init']
                else 1e-2
            )
            rel_step_sz = min(min_step, 1.0 / math.sqrt(param_state['step']))
        param_scale = 1.0
        if param_group['scale_parameter']:
            param_scale = max(param_group['eps2'][1], param_state['RMS'])
        return param_scale * rel_step_sz

    def _get_options(
        self, param_group: typing.Dict[str, typing.Any], param_shape: typing.Tuple[int, ...]
    ) -> typing.Tuple[bool, bool]:
        factored = len(param_shape) >= 2
        use_first_moment = param_group['beta1'] is not None
        return factored, use_first_moment

    def _rms(self, tensor: torch.Tensor) -> float:
        return tensor.norm(2) / (tensor.numel() ** 0.5)

    def _approx_sq_grad(
        self,
        exp_avg_sq_row: torch.Tensor,
        exp_avg_sq_col: torch.Tensor,
        output: torch.Tensor,
    ) -> None:
        r_factor = (
            (exp_avg_sq_row / exp_avg_sq_row.mean(dim=-1))
            .rsqrt_()
            .unsqueeze(-1)
        )
        c_factor = exp_avg_sq_col.unsqueeze(-2).rsqrt()
        torch.mul(r_factor, c_factor, out=output)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        'Adafactor does not support sparse gradients.'
                    )

                state = self.state[p]
                grad_shape = grad.shape

                factored, use_first_moment = self._get_options(
                    group, grad_shape
                )
                # State Initialization
                if len(state) == 0:
                    state['step'] = 0

                    if use_first_moment:
                        # Exponential moving average of gradient values
                        state['exp_avg'] = torch.zeros_like(grad)
                    if factored:
                        state['exp_avg_sq_row'] = torch.zeros(
                            grad_shape[:-1]
                        ).type_as(grad)
                        state['exp_avg_sq_col'] = torch.zeros(
                            grad_shape[:-2] + grad_shape[-1:]
                        ).type_as(grad)
                    else:
                        state['exp_avg_sq'] = torch.zeros_like(grad)

                    state['RMS'] = 0

                state['step'] += 1
                state['RMS'] = self._rms(p.data)
                lr = self._get_lr(group, state)

                beta2t = 1.0 - math.pow(state['step'], group['decay_rate'])
                update = (grad ** 2) + group['eps2'][0]
                if factored:
                    exp_avg_sq_row = state['exp_avg_sq_row']
                    exp_avg_sq_col = state['exp_avg_sq_col']

                    exp_avg_sq_row.mul_(beta2t).add_(
                        update.mean(dim=-1), alpha=1.0 - beta2t
                    )
                    exp_avg_sq_col.mul_(beta2t).add_(
                        update.mean(dim=-2), alpha=1.0 - beta2t
                    )

                    # Approximation of exponential moving average of square
                    # of gradient
                    self._approx_sq_grad(
                        exp_avg_sq_row, exp_avg_sq_col, update
                    )
                    update.mul_(grad)
                else:
                    exp_avg_sq = state['exp_avg_sq']

                    exp_avg_sq.mul_(beta2t).add_(update, alpha=1.0 - beta2t)
                    torch.rsqrt(exp_avg_sq, out=update).mul_(grad)

                update.div_(
                    max(1.0, self._rms(update) / group['clip_threshold'])
                )
                update.mul_(lr)

                if use_first_moment:
                    exp_avg = state['exp_avg']
                    exp_avg.mul_(group['beta1']).add_(
                        update, alpha=1 - group['beta1']
                    )
                    update = exp_avg

                if group['weight_decay'] != 0:
                    p.data.add_(p.data, alpha=-group['weight_decay'] * lr)

                p.data.add_(-update)

        return loss


from moai.utils.torch import get_submodule

import torch
import typing
import toolz

__all__ = ['ModuleParameterSelector']

class ModuleParameterSelector(typing.Callable[[torch.nn.Module], typing.List[torch.Tensor]]):
    def __init__(self,
        keys:       typing.Sequence[str],
        force_grad: bool=True,
    ):
        self.keys = keys
        self.force_grad = force_grad

    def __call__(self, module: torch.nn.Module) -> typing.List[torch.Tensor]:
        params = []
        for key in self.keys: #TODO: check get_submodule in pytorch 1.10
            # split = key.split('.')
            # m = toolz.reduce(getattr, split, module)
            m = get_submodule(module, key)
            params.append(m.parameters())
        parameters = list(toolz.concat(params))
        if self.force_grad:
            for p in parameters:            
                p.requires_grad_(True)
        return parameters


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import typing

__all__ = ['QHAdamW']

#NOTE: modified from https://github.com/iprally/qhadamw-pytorch

class QHAdamW(torch.optim.Optimizer):
    r"""Combines the weight decay decoupling from AdamW with QHAdam:

    - [Decoupled Weight Decay Regularization](https://arxiv.org/pdf/1711.05101.pdf)
    - [Quasi-hyperbolic momentum and Adam for deep learning](https://arxiv.org/pdf/1810.06801.pdf)
    - **Implementation**: [GitHub @ iprally](https://github.com/iprally/qhadamw-pytorch)

    Args:
        params (iterable):
            iterable of parameters to optimize or dicts defining parameter
            groups
        lr (float, optional): learning rate (:math:`\alpha` from the paper)
            (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of the gradient and its square
            (default: (0.995, 0.999))
        nus (Tuple[float, float], optional): immediate discount factors used to
            estimate the gradient and its square
            (default: (0.7, 1.0))
        eps (float, optional): term added to the denominator to improve
            numerical stability
            (default: 1e-8)
        weight_decay (float, optional): weight decay
            (L2 regularization coefficient, times two)
            (default: 0.0)

    Example:
        >>> optimizer = QHAdamW(
        ...     model.parameters(),
        ...     lr=3e-4, nus=(0.8, 1.0), betas=(0.99, 0.999))
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

        QHAdam paper:
    .. _`(Ma and Yarats, 2019)`: https://arxiv.org/abs/1810.06801
        AdamW paper:
    .. _`(Loshchilov and Hutter, 2019)`: https://arxiv.org/abs/1711.05101
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter], 
        lr: float=1e-3, 
        betas: typing.Tuple[float, float]=(0.995, 0.999), 
        nus: typing.Tuple[float, float]=(0.7, 1.0), 
        weight_decay: float=0.0,
        eps: float=1e-8,
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = {"lr": lr, "betas": betas, "nus": nus, "weight_decay": weight_decay, "eps": eps}
        super(QHAdamW, self).__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.

        Args:
            closure (callable, optional):
                A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            beta1, beta2 = group["betas"]
            nu1, nu2 = group["nus"]
            weight_decay = group["weight_decay"]
            eps = group["eps"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                d_p = p.grad.data
                if d_p.is_sparse:
                    raise RuntimeError("QHAdamW does not support sparse gradients")

                param_state = self.state[p]

                # Original QHAdam implementation for weight decay:
                # if weight_decay != 0:
                #    d_p.add_(weight_decay, p.data)

                d_p_sq = d_p.mul(d_p)

                if len(param_state) == 0:
                    param_state["beta1_weight"] = 0.0
                    param_state["beta2_weight"] = 0.0
                    param_state["exp_avg"] = torch.zeros_like(p.data)
                    param_state["exp_avg_sq"] = torch.zeros_like(p.data)

                param_state["beta1_weight"] = 1.0 + beta1 * param_state["beta1_weight"]
                param_state["beta2_weight"] = 1.0 + beta2 * param_state["beta2_weight"]

                beta1_weight = param_state["beta1_weight"]
                beta2_weight = param_state["beta2_weight"]
                exp_avg = param_state["exp_avg"]
                exp_avg_sq = param_state["exp_avg_sq"]

                beta1_adj = 1.0 - (1.0 / beta1_weight)
                beta2_adj = 1.0 - (1.0 / beta2_weight)
                exp_avg.mul_(beta1_adj).add_(1.0 - beta1_adj, d_p)
                exp_avg_sq.mul_(beta2_adj).add_(1.0 - beta2_adj, d_p_sq)

                avg_grad = exp_avg.mul(nu1)
                if nu1 != 1.0:
                    avg_grad.add_(1.0 - nu1, d_p)

                avg_grad_rms = exp_avg_sq.mul(nu2)
                if nu2 != 1.0:
                    avg_grad_rms.add_(1.0 - nu2, d_p_sq)
                avg_grad_rms.sqrt_()
                if eps != 0.0:
                    avg_grad_rms.add_(eps)

                # Original QHAdam implementation:
                # p.data.addcdiv_(-lr, avg_grad, avg_grad_rms)

                # Implementation following AdamW paper:
                p.data.add_(-weight_decay, p.data).addcdiv_(-lr, avg_grad, avg_grad_rms)

        return loss



import torch

class Geodesic(torch.nn.Module):
    r"""Implements the geodesic error function.

    ??? note "Geodesic Loss"
         <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7D%5Cmathcal%7Bd%7D%28R_1%2CR_2%29%20%3D%20%5Carccos%5Cfrac%7Btrace%28R_1R_2^T%29%20-%201%7D%7B2%7D%5Cend%7Bequation%7D"/></p>
    
    ??? cite "Papers"
        [![Paper](https://img.shields.io/static/v1?label=1606.00373&message=On the Continuity of Rotation Representations in Neural Networks&color=1c1ca2&logo=arxiv&style=plastic)](https://arxiv.org/pdf/1812.07035.pdf)

        [![Paper](https://img.shields.io/static/v1?label=1207.6868&message=3D Pose Regression using Convolutional Neural Networks&color=1c1ca2&logo=arxiv&style=plastic)](https://arxiv.org/pdf/1708.05628.pdf)
  
    ???+ example "Configuration"
        === "Main Entry"
            ```yaml
            - model/supervision/losses/object_pose: geodesic
            ```
        === "Parameters"
            _The geodesic loss has no parameters currently._
        === "Graph"
            ```yaml
            model:
              supervision:
                losses:
                  geodesic:
                    gt: [groundtruth_tensor_name] # the name of the groundtruth tensor
                    pred: [prediction_tensor_name] # the name of the predicted tensor
                    out: [geodesic_loss] # optional, will be 'geodesic' if omitted
            ```
    
  
    !!! important
        Both matrices should be orthogonal.

    """
    def __init__(self):
        super(Geodesic, self).__init__()

    def forward(self,
        pred: torch.Tensor, # b x 3 x 3
        gt: torch.Tensor, # b x 3 x 3        
    ) -> torch.Tensor:        
        return _compute_geodesic_loss(gt, pred)

def _compute_geodesic_loss(gt_r_matrix, out_r_matrix):
    batch = gt_r_matrix.shape[0]
    m = torch.bmm(gt_r_matrix, out_r_matrix.transpose(1, 2)) # [B, 3, 3]
    cos = (m[:, 0, 0] + m[:, 1, 1] + m[:, 2, 2] - 1.0) / 2.0
    cos = torch.min(cos, torch.ones(batch).to(gt_r_matrix))
    torch.max(cos, torch.ones(batch).to(gt_r_matrix) * -1.0)
    theta = torch.acos(cos)
    return theta


from moai.monads.utils import (
    spatial_dim_list,
    flatten_spatial_dims,
    spatial_dims,
    expand_spatial_dims,
)

import torch
import functools

__all__ = ["CoordinateDecoding"]

# NOTE: generic helper functions should go to utils? 
def unravel_indices(
    indices: torch.Tensor,
    shape: torch.Tensor
) -> torch.LongTensor:
    r"""Converts flat indices into unraveled coordinates in a target shape.

    Args:
        indices: A tensor of (flat) indices, (*, N).
        shape: The targeted shape, (D,).

    Returns:
        The unraveled coordinates, (*, N, D).
    """

    coord = []
    for i, dim in enumerate(reversed(shape)):
        coord.append(indices % dim)
        indices = indices // dim
    coord = torch.stack(coord[::-1], dim=-1).long()
    return coord

class CoordinateDecoding(torch.nn.Module):

    @staticmethod
    def _argmax(
            grid: torch.Tensor,     # [B, (S)VU, (D), H, W], order is (S)VU not UV(S), y coord first channel
            heatmaps: torch.Tensor  # [B, K, (D), H, W], with its value across the spatial dimensions summing to unity
        ) -> torch.Tensor:          # [B, K, UV(S) or (S)VU]
            channels = heatmaps.shape[1]        
            dims = spatial_dims(heatmaps)
            coords = []
            for i in range(channels): #TODO: refactor for loop
                heatmap = heatmaps[:, i, ...].unsqueeze(1)
                flattened_heatmap = flatten_spatial_dims(heatmap).squeeze(1)
                indices = torch.argmax(flattened_heatmap, dim=1)
                raw_m = unravel_indices(indices, dims)
                coord = torch.zeros_like(raw_m).float()
                for b in range(coord.shape[0]):                
                    for j in range(len(dims)):
                        coord[b, j] = grid[b, j][tuple(raw_m[b])]
                coords.append(coord)
            return torch.stack(coords, dim=1)

    @staticmethod
    def _argmax_sub(
            grid: torch.Tensor,         # [B, (S)VU, (D), H, W], order is (S)VU not UV(S), y coord first channel
            heatmaps: torch.Tensor,     # [B, K, (D), H, W], with its value across the spatial dimensions summing to unity
            sub: float                  # sub-pixel shifting to second maximal activation
        ) -> torch.Tensor:              # [B, K, UV(S) or (S)VU]
            channels = heatmaps.shape[1]        
            dims = spatial_dims(heatmaps)
            argmax_dims = spatial_dim_list(heatmaps)
            coords = []
            for i in range(channels): #TODO: refactor for loop
                heatmap = heatmaps[:, i, ...].unsqueeze(1)
                flattened_heatmap = flatten_spatial_dims(heatmap).squeeze(1)
                indices = torch.topk(flattened_heatmap, 2, dim=1)[1]
                
                raw_m = unravel_indices(indices[:, 0], dims)
                m = torch.zeros_like(raw_m).float()
                for b in range(m.shape[0]):                
                    for j in range(len(dims)):
                        m[b, j] = grid[b, j][tuple(raw_m[b])]

                raw_s = unravel_indices(indices[:, 1], dims)
                s = torch.zeros_like(raw_s).float()
                for b in range(s.shape[0]):                
                    for j in range(len(dims)):
                        s[b, j] = grid[b, j][tuple(raw_s[b])]

                coord = m + sub * (s - m) * (grid.amax(dim=argmax_dims) - \
                    grid.amin(dim=argmax_dims)) / (torch.linalg.norm(s - m) * dims)
                coords.append(coord)
            return torch.stack(coords, dim=1)

    def __init__(self,
        mode:           str="argmax", #NOTE: to support argmax, standard
        flip:           bool=True, # flip order of coordinates as originated from the grid's order
        sub:            float=0.25
    ):
        super(CoordinateDecoding, self).__init__()
        self.mode = mode
        self.flip = flip
        self.sub = sub
        self.branch = functools.partial(self._argmax_sub, sub=sub)\
            if mode == 'standard' else functools.partial(self._argmax)

                
    def forward(self, 
        grid: torch.Tensor,     # coordinates grid tensor of C coordinates
        heatmaps: torch.Tensor, # spatial probability tensor of K keypoints
    ) -> torch.Tensor:            
            coords = self.branch(grid, heatmaps)
            return torch.flip(coords, dims=[2]) if self.flip else coords


import torch
import functools

__all__ = ["KL"]

class KL(torch.nn.KLDivLoss):
    def __init__(self,
        is_input_log:   bool=False,
        is_target_log:  bool=False,
    ):
        super(KL, self).__init__(reduction='none', log_target=is_target_log)
        self.is_input_log = is_input_log

    def forward(self, 
        pred: torch.Tensor,
        gt: torch.Tensor,        
    ) -> torch.Tensor:
        #NOTE: is eps numerical stabilization required?
        return super(KL, self).forward(pred if self.is_input_log else pred.log(), gt)


import math
import torch
import torch.nn as nn
import typing

__all__ = ['Yogi']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/yogi.py

class Yogi(torch.optim.Optimizer):
    r"""Implements Yogi Optimizer Algorithm.

    - **Paper**: [Adaptive methods for Nonconvex Optimization](https://papers.nips.cc/paper/8186-adaptive-methods-for-nonconvex-optimization.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)    

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate (default: 1e-2)
        betas: coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps: term added to the denominator to improve
            numerical stability (default: 0.001)
        initial_accumulator: initial values for first and
            second moments (default: 1e-6)
        weight_decay: weight decay (L2 penalty) (default: 0)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.Yogi(model.parameters(), lr=0.01)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ https://papers.nips.cc/paper/8186-adaptive-methods-for-nonconvex-optimization  # noqa

    Note:
        Reference code: https://github.com/4rtemi5/Yogi-Optimizer_Keras
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-2,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-3,
        initial_accumulator: float=1e-6,
        weight_decay: float=0,
    ) -> None:
        if lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))
        if eps < 0.0:
            raise ValueError('Invalid epsilon value: {}'.format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(
                'Invalid beta parameter at index 0: {}'.format(betas[0])
            )
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(
                'Invalid beta parameter at index 1: {}'.format(betas[1])
            )
        if weight_decay < 0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )

        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            initial_accumulator=initial_accumulator,
            weight_decay=weight_decay,
        )
        super(Yogi, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        'Yogi does not support sparse gradients, '
                        'please consider SparseAdam instead'
                    )

                state = self.state[p]

                # State initialization
                # Followed from official implementation in tensorflow addons:
                # https://github.com/tensorflow/addons/blob/master/tensorflow_addons/optimizers/yogi.py#L118 # noqa
                # For more details refer to the discussion:
                # https://github.com/jettify/pytorch-optimizer/issues/77
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = nn.init.constant_(
                        torch.empty_like(
                            p.data, memory_format=torch.preserve_format
                        ),
                        group['initial_accumulator'],
                    )
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = nn.init.constant_(
                        torch.empty_like(
                            p.data, memory_format=torch.preserve_format
                        ),
                        group['initial_accumulator'],
                    )

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                if group['weight_decay'] != 0:
                    grad = grad.add(p.data, alpha=group['weight_decay'])

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                grad_squared = grad.mul(grad)

                exp_avg_sq.addcmul_(
                    torch.sign(exp_avg_sq - grad_squared),
                    grad_squared,
                    value=-(1 - beta2),
                )

                denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(
                    group['eps']
                )
                step_size = group['lr'] / bias_correction1
                p.data.addcdiv_(exp_avg, denom, value=-step_size)

        return loss


from collections.abc import Callable, Iterator
from moai.monads.execution.cascade import _create_accessor

import numpy as np
import npstreams
import typing
import torch
import logging

log = logging.getLogger(__name__)

__all__ = ['Pairwise']

class Pairwise(Callable):        
    def __init__(self,
        key:        typing.Union[str, typing.Sequence[str]],
        pair:      typing.Sequence[typing.Tuple[int, int]],
    ) -> None:
        super().__init__()
        self.names = [key] if isinstance(key, str) else list(key)
        self.data = { }
        self.keys = [_create_accessor(k) for k in self.names]
        self.pairs = pair
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for n, k, p in zip(self.names, self.keys, self.pairs):
            t = k(tensors)
            id = f"{n}_{p[0]}-{p[1]}"
            if id not in self.data:
                self.data[id] = t[:, p]
            else:
                self.data[id] = torch.cat([self.data[id], t[:, p]])
        tensors.update(self.data)


from moai.monads.utils import flatten_spatial_dims

import torch

__all__ = ["SpatialSoftmax"]

#NOTE: see "FlowCap: 2D Human Pose from Optical Flow" for sharpening
#NOTE: see https://timvieira.github.io/blog/post/2014/02/11/exp-normalize-trick/

class SpatialSoftmax(torch.nn.Module):
    def __init__(self,
        temperature:    float=1.0,  # smoothen out the output by premultiplying input
        alpha:          float=1.0,  # sharpen output
        normalize:      bool=False, # normalize output
    ):
        super(SpatialSoftmax, self).__init__()
        self.temp = temperature
        self.alpha = alpha
        self.normalize = normalize
        #TODO: add inplace version / flag

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:        
        reduced = flatten_spatial_dims(tensor)
        if self.temp != 1.0:
            reduced = reduced * self.temp
        if self.alpha != 1.0:
            reduced = reduced ** self.alpha
        if self.normalize:
            reduced = reduced / reduced.flatten(2).sum(-1)
        softmaxed = torch.nn.functional.softmax(reduced, dim=-1)
        return softmaxed.view_as(tensor)


from moai.parameters.optimization.optimizers.swa import SWA as Outer

import torch
import omegaconf.omegaconf
import hydra.utils as hyu
import typing

__all__ = ['SWA']

#NOTE: modified from https://github.com/alphadl/lookahead.pytorch

class SWA(object):
    """Implements Stochastic Weight Averaging (SWA).

        - **Paper**: [Averaging Weights Leads to Wider Optima and Better Generalization](https://arxiv.org/pdf/1803.05407.pdf)
        - **Implementation**: [GitHub @ pytorch](https://github.com/pytorch/contrib/tree/master/torchcontrib)

    """
    def __init__(self,
        parameters: typing.Iterator[torch.nn.Parameter],
        optimizers: omegaconf.DictConfig,        
        swa_start=None,
        swa_freq=None,
        swa_lr=None
    ):
        self.optimizers = [
            Outer(
                optimizer=hyu.instantiate(optimizers[0], parameters), 
                swa_start=swa_start, swa_lr=swa_lr, swa_freq=swa_freq,
            )
        ]



from moai.monads.utils.spatial import expand_spatial_dims

import torch

__all__ = ["Berhu"]

class Berhu(torch.nn.Module):
    r"""Implements the (adaptive) berHu error function.

    ??? note "Berhu Loss"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7D%5Cmathcal%7BB%7D%28x%29%20%3D%20%5Cbegin%7Bcases%7D%20|x|%20%26%20|x|%20%5Cleq%20c%2C%20%5C%5C%20%5Cfrac%7Bx^2%20%2B%20c^2%7D%7B2c%7D%20%26%20|x|%20%3E%20c.%20%5C%5C%20%5Cend%7Bcases%7D%20%5Cend%7Bequation%7D"/></p>

    ??? note "Adaptive Threshold"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%20c%20%3D%20%5Cfrac%7B1%7D%7B5%7D%20%5Cmax_i%28|%5Ctilde%7By%7D_i%20-%20y_i|%29"/></p>

    ??? cite "Papers"
        [![Paper](https://img.shields.io/static/v1?label=1606.00373&message=Deeper Depth Prediction with Fully Convolutional Residual Networks&color=1c1ca2&logo=arxiv&style=plastic)](https://arxiv.org/pdf/1606.00373.pdf)

        [![Paper](https://img.shields.io/static/v1?label=1207.6868&message=The BerHu penalty and the grouped effect&color=1c1ca2&logo=arxiv&style=plastic)](https://arxiv.org/pdf/1207.6868.pdf)
 
    ???+ example "Configuration"
        === "Main Entry"
            ```yaml
            - model/supervision/losses/regression/robust: berhu
            ```
        === "Parameters"
            ```yaml            
            threshold: 1.0
            adaptive: false # toggle adaptive threshold, ignores given threshold
            image_wide: false # adaptive threshold calculated per batch item (true) or per batch (false)
            ```
        === "DML"
            ```yaml
            model:
              supervision:
                losses:
                  berhu:
                    threshold: 1.0
                    adaptive: false
                    image_wide: false
            ```
        === "Graph"
            ```yaml
            model:
              supervision:
                losses:
                  berhu:
                    gt: [groundtruth_tensor_name] # the name of the groundtruth tensor
                    pred: [prediction_tensor_name] # the name of the predicted tensor
                    out: [berhu_loss] # optional, will be 'berhu' if omitted
            ```
       
    Arguments:
        threshold (float, optional):
            Sets the threshold that switches between the L1 and L2 loss.       
        adaptive (boolean, optional):
            If true, the threshold is calculated dynamically for each batch. **Default value: False.**
        image_wide (boolean, optional): 
            If true, the adaptive threshold is calculated for each batch item. **Default value: False.**        

    !!! important
        Can be helpful when dealing with heavy-tailed distribution of predicted values.

    ??? info "Repositories"
        [![Repo1](https://github-readme-stats.vercel.app/api/pin/?username=iro-cp&repo=FCRN-DepthPrediction&hide_border=true&title_color=1c1ca2&show_owner=true)](https://github.com/iro-cp/FCRN-DepthPrediction){: align=left }
        [![Repo2](https://github-readme-stats.vercel.app/api/pin/?username=dontLoveBugs&repo=FCRN_pytorch&hide_border=true&title_color=1c1ca2&show_owner=true)](https://github.com/dontLoveBugs/FCRN_pytorch){: align=right }

    *[berHu]: Reverse Huber

    """
    def __init__(self,
        threshold:      float,
        adaptive:       bool=False,
        image_wide:     bool=False,
    ):
        super(Berhu, self).__init__()
        self.threshold = threshold
        self.adaptive = adaptive
        self.image_wide = image_wide

    def _get_max(self, error: torch.Tensor) -> torch.Tensor:
        if self.image_wide:
            values, indices = torch.max(error.view(error.shape[0], -1), dim=1)
            expanded = expand_spatial_dims(values, error)
            if error.shape[1] != 1:
                expanded = expanded.expand(error.shape[0], error.shape[1], 1)
            return expanded
        else:
            return torch.max(error)

    def forward(self, 
        pred:       torch.Tensor,
        gt:         torch.Tensor,        
        weights:    torch.Tensor=None,       # float tensor
        mask:       torch.Tensor=None,       # byte tensor
    ) -> torch.Tensor:        
        L1 = torch.abs(gt - pred)
        if weights is not None:
            L1 = L1 * weights
        if mask is not None:
            L1 = L1[mask]
        threshold = self.threshold if not self.adaptive else 0.2 * self._get_max(L1)
        if len(threshold.shape) > 1:
            L1 = L1.view(*threshold.shape[:2], -1)
        berhu = torch.where(
            L1 <= threshold,
            L1,
            (L1 ** 2 + threshold ** 2) / (2.0 * threshold)
        )
        return berhu.view_as(gt)


from human_body_prior.models.vposer_model import VPoser

import toolz
import torch
import logging
import typing
import os
import glob

log = logging.getLogger(__name__)

__all__ = ["VPoser2"]

class VPoser2(typing.Callable[[torch.nn.Module], None]):
    def __init__(self, 
        ckpt:       str,        
        cache:      bool=False,
    ):
        if os.path.isdir(ckpt):
            self.filename = glob.glob(os.path.join(ckpt, 'snapshots', '*.ckpt'))[-1]
        else:
            self.filename = ckpt
        if cache:
            self.state_dict = torch.load(
                self.filename, map_location=lambda storage, loc: storage
            )['state_dict']
        
    def __call__(self,
        module: torch.nn.Module
    ) -> None:        
        if isinstance(module, VPoser):
            state = self.state_dict if hasattr(self, 'state_dict') else torch.load(
                self.filename, map_location=lambda storage, loc: storage
            )['state_dict']
            module.load_state_dict(
                toolz.keymap(lambda k: k.replace('vp_model.', ''), state),
                strict=False
            )
            log.info(f"Initialized VPoser2 from {self.filename}")